#!/usr/bin/python
# -*- coding: utf8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.


import shelve
import sys
import datetime
import decimal
import os
from .utils import leer, escribir, leer_dbf, guardar_dbf, N, A, I, json, BaseWS, inicializar_y_capturar_excepciones, get_install_dir
from . import utils
from fpdf import Template
from pysimplesoap.client import SoapFault
import pprint
import traceback
"""Módulo para obtener código de autorización electrónica (CAE) para
Liquidación Sector Pecuario (hacienda/carne) del web service WSLSP de AFIP
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2016 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.07a"

LICENCIA = """
wslsp.py: Interfaz para generar Código de Autorización Electrónica (CAE) para
          Liquidación Sector Pecuario (LspService)
Copyright (C) 2016 Mariano Reingart reingart@gmail.com
http://www.sistemasagiles.com.ar/trac/wiki/LiquidacionSectorPecuario

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo respetando la licencia GPLv3.

Para información adicional sobre garantía, soporte técnico comercial
e incorporación/distribución en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

AYUDA = """
Opciones:
  --ayuda: este mensaje

  --debug: modo depuración (detalla y confirma las operaciones)
  --formato: muestra el formato de los archivos de entrada/salida
  --prueba: genera y autoriza una liquidación de prueba (no usar en producción!)
  --xml: almacena los requerimientos y respuestas XML (depuración)
  --json: utilizar formato json para el archivo de intercambio
  --dummy: consulta estado de servidores

  --autorizar: Autorizar Liquidación Única (generarLiquidacion)
  --ajustar: Ajuste Físico/Monetario/Financiero, Credito/Debito (generarAjuste)
  --ult: Consulta el último número de orden registrado en AFIP
         (consultarUltimoComprobanteXPuntoVenta)
  --consultar: Consulta una liquidación registrada en AFIP
         (consultarLiquidacionXNroComprobante / consultarLiquidacionXCAE)

  --provincias: obtiene el listado de provincias (código/descripción)
  --localidades: obtiene el listado de localidades para una provincia
  --tributos: obtiene el listado de los tipos de tributos
  --gastos: obtiene el listado de los tipos de gastos
  --puntosventa: obtiene el listado de puntos de venta habilitados

Ver wslsp.ini para parámetros de configuración (URL, certificados, etc.)"
"""


# importo funciones compartidas:


WSDL = "https://fwshomo.afip.gov.ar/wslsp/LspService?wsdl"
#WSDL = "https://serviciosjava.afip.gov.ar/wslsp/LspService?wsdl"

DEBUG = False
XML = False
CONFIG_FILE = "wslsp.ini"
HOMO = False


class WSLSP(BaseWS):
    "Interfaz para el WebService de Liquidación Única Mensual (lechería)"
    _public_methods_ = ['Conectar', 'Dummy', 'SetTicketAcceso', 'DebugLog',
                        'AutorizarLiquidacion',
                        'CrearLiquidacion', 'AgregarFrigorifico',
                        'AgregarEmisor', 'AgregarReceptor', 'AgregarOperador',
                        'AgregarDTE',
                        'AgregarItemDetalle', 'AgregarCompraAsociada',
                        'AgregarGasto', 'AgregarTributo', 'AgregarGuia',
                        'ConsultarLiquidacion', 'ConsultarUltimoComprobante',
                        'CrearAjuste', 'AgregarComprobanteAAjustar',
                        'AgregarItemDetalleAjuste',
                        'AgregarAjusteMonetario', 'AgregarAjusteFisico',
                        'AgregarAjusteFinanciero',
                        'AjustarLiquidacion',
                        'LeerDatosLiquidacion',
                        'ConsultarOperaciones',
                        'ConsultarTiposComprobante',
                        'ConsultarTiposLiquidacion',
                        'ConsultarCategorias', 'ConsultarMotivos',
                        'ConsultarRazas', 'ConsultarCortes',
                        'ConsultarCaracteresParticipante',
                        'ConsultarGastos', 'ConsultarTributos',
                        'ConsultarPuntosVentas',
                        'ConsultarProvincias', 'ConsultarLocalidades',
                        'AnalizarXml', 'ObtenerTagXml', 'LoadTestXML',
                        'SetParametros', 'SetParametro', 'GetParametro',
                        ]
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
                      'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
                      'Excepcion', 'ErrCode', 'ErrMsg', 'LanzarExcepciones', 'Errores',
                      'XmlRequest', 'XmlResponse', 'Version', 'Traceback', 'InstallDir',
                      'CAE', 'NroComprobante', 'FechaComprobante',
                      'NroCodigoBarras', 'FechaVencimientoCae', 'FechaProcesoAFIP',
                      'ImporteBruto', 'ImporteIVASobreBruto', '',
                      'ImporteTotalGastos', 'ImporteIVASobreGastos',
                      'ImporteTotalTributos', 'ImporteTotalNeto',
                      ]
    _reg_progid_ = "WSLSP"
    _reg_clsid_ = "{9750BBD4-FBC3-4FE7-8DE5-E193667D6813}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    LanzarExcepciones = False
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')

    def inicializar(self):
        BaseWS.inicializar(self)
        self.AppServerStatus = self.DbServerStatus = self.AuthServerStatus = None
        self.errores = []
        self.CAE = ""
        self.NroComprobante = self.FechaComprobante = self.FechaProcesoAFIP = ''
        self.NroCodigoBarras = self.FechaVencimientoCae = None
        self.ImporteBruto = self.ImporteIVASobreBruto = None
        self.ImporteTotalGastos = self.ImporteIVASobreGastos = None
        self.ImporteTotalTributos = self.ImporteTotalNeto = None
        self.datos = {}

    @inicializar_y_capturar_excepciones
    def Conectar(self, cache=None, url="", proxy="", wrapper="", cacert=None, timeout=60):
        "Establecer la conexión a los servidores de la AFIP"
        # llamo al constructor heredado:
        ok = BaseWS.Conectar(self, cache, url, proxy, wrapper, cacert, timeout)
        if False and ok:
            # corrijo ubicación del servidor (puerto htttp 80 en el WSDL)
            location = self.client.services['LspService']['ports']['LumEndPoint']['location']
            if location.startswith("http://"):
                print("Corrigiendo WSDL ...", location, end=' ')
                location = location.replace("http://", "https://").replace(":80", ":443")
                self.client.services['LspService']['ports']['LspEndPoint']['location'] = location
                print(location)
        return ok

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        errores = []
        if 'errores' in ret:
            errores.extend(ret['errores'])
        if errores:
            self.Errores = ["%(codigo)s: %(descripcion)s" % err['error'][0]
                            for err in errores]
            self.errores = [
                {'codigo': err['error'][0]['codigo'],
                 'descripcion': err['error'][0]['descripcion'].replace("\n", "")
                 .replace("\r", "")}
                for err in errores]
            self.ErrCode = ' '.join(self.Errores)
            self.ErrMsg = '\n'.join(self.Errores)

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = self.client.dummy()['respuesta']
        self.AppServerStatus = str(results['appserver'])
        self.DbServerStatus = str(results['dbserver'])
        self.AuthServerStatus = str(results['authserver'])
        return True

    @inicializar_y_capturar_excepciones
    def CrearLiquidacion(self, cod_operacion, fecha_cbte, fecha_op, cod_motivo,
                         cod_localidad_procedencia, cod_provincia_procedencia,
                         cod_localidad_destino, cod_provincia_destino,
                         lugar_realizacion=None,
                         fecha_recepcion=None, fecha_faena=None,
                         datos_adicionales=None, **kwargs):
        "Inicializa internamente los datos de una liquidación para autorizar"
        # creo el diccionario con los campos generales de la liquidación:
        liq = {'fechaComprobante': fecha_cbte, 'fechaOperacion': fecha_op,
               'lugarRealizacion': lugar_realizacion,
               'codMotivo': cod_motivo,
               'codLocalidadProcedencia': cod_localidad_procedencia,
               'codProvinciaProcedencia': cod_provincia_procedencia,
               'codLocalidadDestino': cod_localidad_destino,
               'codProvinciaDestino': cod_provincia_destino,
               'fechaRecepcion': fecha_recepcion,
               'fechaFaena': fecha_faena,
               }
        self.solicitud = dict(codOperacion=cod_operacion,
                              emisor={}, receptor={},
                              datosLiquidacion=liq,
                              itemDetalleLiquidacion=[],
                              guia=[], dte=[],
                              tributo=[], gasto=[],
                              datosAdicionales=datos_adicionales,
                              )
        return True

    @inicializar_y_capturar_excepciones
    def AgregarFrigorifico(self, cuit, nro_planta):
        "Agrego el frigorifico a la liquidacíon (opcional)."
        frig = {'cuit': cuit, 'nroPlanta': nro_planta}
        self.solicitud['datosLiquidacion']['frigorifico'] = frig
        return True

    @inicializar_y_capturar_excepciones
    def AgregarEmisor(self, tipo_cbte, pto_vta, nro_cbte, cod_caracter=None,
                      fecha_inicio_act=None, iibb=None, nro_ruca=None,
                      nro_renspa=None, cuit_autorizado=None, **kwargs):
        "Agrego los datos del emisor a la liq."
        # cod_caracter y fecha_inicio_act no es requerido para ajustes
        d = {'tipoComprobante': tipo_cbte, 'puntoVenta': pto_vta,
             'nroComprobante': nro_cbte,
             'codCaracter': cod_caracter,
             'fechaInicioActividades': fecha_inicio_act,
             'iibb': iibb,
             'nroRUCA': nro_ruca,
             'nroRenspa': nro_renspa,
             'cuitAutorizado': cuit_autorizado}
        self.solicitud['emisor'].update(d)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarReceptor(self, cod_caracter, **kwargs):
        "Agrego los datos del receptor a la liq."
        d = {'codCaracter': cod_caracter}
        self.solicitud['receptor'].update(d)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarOperador(self, cuit, iibb=None, nro_ruca=None, nro_renspa=None,
                        cuit_autorizado=None, **kwargs):
        "Agrego los datos del operador a la liq."
        d = {'cuit': cuit,
             'iibb': iibb,
             'nroRUCA': nro_ruca,
             'nroRenspa': nro_renspa,
             'cuitAutorizado': cuit_autorizado}
        self.solicitud['receptor']['operador'] = d
        return True

    @inicializar_y_capturar_excepciones
    def AgregarItemDetalle(self, cuit_cliente, cod_categoria, tipo_liquidacion,
                           cantidad, precio_unitario, alicuota_iva, cod_raza,
                           cantidad_cabezas=None, nro_tropa=None,
                           cod_corte=None, cantidad_kg_vivo=None,
                           precio_recupero=None, detalle_raza=None,
                           nro_item=None,
                           **kwargs):
        "Agrega el detalle de item de la liquidación"
        d = {'cuitCliente': cuit_cliente,
             'codCategoria': cod_categoria,
             'tipoLiquidacion': tipo_liquidacion,
             'cantidad': cantidad,
             'precioUnitario': precio_unitario,
             'alicuotaIVA': alicuota_iva,
             'cantidadCabezas': cantidad_cabezas,
             'raza': {'codRaza': cod_raza, 'detalle': detalle_raza},
             'nroTropa': nro_tropa,
             'codCorte': cod_corte,
             'cantidadKgVivo': cantidad_kg_vivo,
             'precioRecupero': precio_recupero,
             'liquidacionCompraAsociada': [],
             'nroItem': nro_item,
             }
        self.solicitud['itemDetalleLiquidacion'].append(d)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarCompraAsociada(self, tipo_cbte, pto_vta, nro_cbte, cant_asoc, nro_item):
        "Agrega la información referente a la liquidación compra asociada"
        d = {'tipoComprobante': tipo_cbte,
             'puntoVenta': pto_vta,
             'nroComprobante': nro_cbte,
             'nroItem': nro_item,
             'cantidadAsociada': cant_asoc}
        if 'itemDetalleLiquidacion' in self.solicitud:
            item_liq = self.solicitud['itemDetalleLiquidacion'][-1]
            item_liq['liquidacionCompraAsociada'].append(d)
        else:
            item_liq = self.solicitud['itemDetalleAjusteLiquidacion'][-1]
            item_liq['ajusteCompraAsociada'].append(d)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarGasto(self, cod_gasto, descripcion=None, base_imponible=None,
                     alicuota=None, importe=None, alicuota_iva=None,
                     tipo_iva_nulo=None):
        "Agrega la información referente a los gastos de la liquidación"
        # WSLSPv1.4.1: tipo_iva_nulo debe ser NG, NA, EX: Exento.
        if alicuota_iva == 0:
            alicuota_iva = None         # sólo acepta [10.5, 21.0]
        elif alicuota_iva:
            tipo_iva_nulo = None
        gasto = {'codGasto': cod_gasto, 'descripcion': descripcion,
                 'baseImponible': base_imponible, 'alicuota': alicuota,
                 'importe': importe, 'alicuotaIVA': alicuota_iva,
                 'tipoIVANulo': tipo_iva_nulo,
                 }
        if 'ajusteFinanciero' in self.solicitud:
            self.solicitud['ajusteFinanciero']['gasto'].append(gasto)
        else:
            self.solicitud['gasto'].append(gasto)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarTributo(self, cod_tributo, descripcion=None,
                       base_imponible=None, alicuota=None, importe=None):
        "Agrega la información referente a los tributos de la liquidación"
        trib = {'codTributo': cod_tributo, 'descripcion': descripcion,
                'baseImponible': base_imponible, 'alicuota': alicuota,
                'importe': importe}
        if 'ajusteFinanciero' in self.solicitud:
            self.solicitud['ajusteFinanciero']['tributo'].append(trib)
        else:
            self.solicitud['tributo'].append(trib)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarDTE(self, nro_dte, nro_renspa=None):
        "Agrega la información referente a DTE (multiples)"
        d = {"nroDTE": nro_dte, "nroRenspa": nro_renspa}
        self.solicitud['dte'].append(d)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarGuia(self, nro_guia):
        "Agrega la información referente a las guías (multiples)"
        self.solicitud['guia'].append({"nroGuia": nro_guia})
        return True

    @inicializar_y_capturar_excepciones
    def AutorizarLiquidacion(self):
        "Generar o ajustar una liquidación única y obtener del CAE"
        # limpio los elementos que no correspondan por estar vacios:
        for campo in ["guia", "dte", "gasto", "tributo"]:
            if campo in self.solicitud and not self.solicitud[campo]:
                del self.solicitud[campo]
        for item in self.solicitud['itemDetalleLiquidacion']:
            if not item.get("liquidacionCompraAsociada", True):
                del item["liquidacionCompraAsociada"]
        # llamo al webservice:
        ret = self.client.generarLiquidacion(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
            solicitud=self.solicitud,
        )
        # analizo la respusta
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        self.AnalizarLiquidacion(ret)
        return True

    def AnalizarLiquidacion(self, liq):
        "Método interno para analizar la respuesta de AFIP"
        # proceso los datos básicos de la liquidación (devuelto por consultar):
        cab = liq.get('cabecera')
        if cab:
            self.CAE = str(cab.get('cae', ''))
            self.FechaVencimientoCae = str(cab.get('fechaVencimientoCae', ''))
            self.FechaProcesoAFIP = str(cab.get('fechaProcesoAFIP', ''))
            self.NroCodigoBarras = cab.get('nroCodigoBarras')
            datos = liq.get('datosLiquidacion', {})
            self.FechaComprobante = str(datos.get('fechaComprobante', ''))
            emisor = liq.get('emisor', {})
            self.NroComprobante = emisor.get('nroComprobante')
            tot = liq.get('resumenTotales', {})
            self.ImporteBruto = tot.get('importeBruto')
            self.ImporteTotalGastos = tot.get('importeTotalGastos')
            self.ImporteTotalTributos = tot.get('importeTotalTributos')
            self.ImporteTotalNeto = tot.get('importeTotalNeto')
            self.ImporteIVASobreBruto = tot.get('importeIVASobreBruto')
            self.ImporteIVASobreGastos = tot.get('importeIVASobreGastos')
            receptor = liq.get('receptor', {})

            # parámetros de salida:
            self.params_out = dict(
                tipo_cbte=emisor.get('tipoComprobante'),
                pto_vta=emisor.get('puntoVenta'),
                nro_cbte=emisor.get('nroComprobante'),
                fecha=datos.get('fechaComprobante'),
                cae=cab.get('cae'),
                emisor=dict(
                    razon_social=emisor.get('razonSocial'),
                    domicilio_punto_venta=emisor.get('domicilioPuntoVenta'),
                ),
                receptor=dict(
                    nombre=receptor.get('nombre'),
                    domicilio=receptor.get('domicilio'),
                ),
                bruto=tot.get('importeBruto'),
                iva_bruto=tot.get('importeIVASobreBruto'),
                iva_gastos=tot.get('importeIVASobreGastos'),
                total_neto=tot.get('importeTotalNeto'),
                total_tributos=tot.get('importeTotalTributos'),
                total_gastos=tot.get('importeTotalGastos'),
                gasto=[],
                guia=[],
                tributo=[],
                pdf=liq.get('pdf'),
            )
            for ret in liq.get('gasto', []):
                self.params_out['gasto'].append(dict(
                    cod_gasto=ret['codGasto'],
                    importe=ret['importe'],
                ))
            for trib in liq.get('tributo', []):
                self.params_out['tributo'].append(dict(
                    descripcion=trib.get('descripcion', ""),
                    base_imponible=trib.get('baseImponible'),
                    alicuota=trib.get('alicuota'),
                    codigo=trib['codTributo'],
                    importe=trib['importe'],
                ))
            # analizar datos del ajuste generado:
            ajuste = liq.get('ajuste', {})
            if ajuste:
                self.params_out.update(dict(
                    tipo_ajuste=ajuste['tipoAjuste'],
                    modo_ajuste=ajuste['modoAjuste'],
                ))
                ajustado = ajuste.get('comprobanteAjustado')
                if ajustado:
                    self.params_out.update(dict(
                        cbte_ajuste=dict(
                            tipo_cbte=ajustado['tipoComprobante'],
                            pto_vta=ajustado['puntoVenta'],
                            nro_cbte=ajustado['nroComprobante'],
                        )
                    ))
            if DEBUG:
                import pprint
                pprint.pprint(self.params_out)
        self.params_out['errores'] = self.errores

    @inicializar_y_capturar_excepciones
    def ConsultarLiquidacion(self, tipo_cbte=None, pto_vta=None, nro_cbte=None,
                             cae=None, cuit_comprador=None, pdf="liq.pdf"):
        "Consulta una liquidación por No de Comprobante o CAE"
        if cae:
            ret = self.client.consultarLiquidacionPorCae(
                auth={
                    'token': self.Token, 'sign': self.Sign,
                    'cuit': self.Cuit, },
                solicitud={
                    'cae': cae,
                    'pdf': pdf and True or False,
                },
            )
        else:
            ret = self.client.consultarLiquidacionPorNroComprobante(
                auth={
                    'token': self.Token, 'sign': self.Sign,
                    'cuit': self.Cuit, },
                solicitud={
                    'cuitComprador': cuit_comprador,
                    'puntoVenta': pto_vta,
                    'nroComprobante': nro_cbte,
                    'tipoComprobante': tipo_cbte,
                    'pdf': pdf and True or False,
                },
            )
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        self.AnalizarLiquidacion(ret)
        # guardo el PDF si se indico archivo y vino en la respuesta:
        if pdf and 'pdf' in ret:
            open(pdf, "wb").write(ret['pdf'])
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarUltimoComprobante(self, tipo_cbte=151, pto_vta=1):
        "Consulta el último No de Comprobante registrado"
        ret = self.client.consultarUltimoNroComprobantePorPtoVta(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
            solicitud={
                'puntoVenta': pto_vta,
                'tipoComprobante': tipo_cbte},
        )
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        self.NroComprobante = ret['nroComprobante']
        return True

    @inicializar_y_capturar_excepciones
    def CrearAjuste(self, tipo_ajuste, fecha_cbte, datos_adicionales=None, **kwargs):
        "Inicializa internamente los datos de una liquidación para ajustar"
        # creo el diccionario con los campos generales de la liquidación:
        self.solicitud = dict(tipoAjuste=tipo_ajuste,
                              fechaComprobante=fecha_cbte,
                              emisor={},
                              itemDetalleAjusteLiquidacion=[],
                              ajusteFinanciero={},
                              datosAdicionales=datos_adicionales,
                              )
        return True

    @inicializar_y_capturar_excepciones
    def AgregarComprobanteAAjustar(self, tipo_cbte, pto_vta, nro_cbte):
        "Agrega comprobante a ajustar"
        cbte = dict(tipoComprobante=tipo_cbte, puntoVenta=pto_vta, nroComprobante=nro_cbte)
        self.solicitud['emisor']["comprobanteAAjustar"] = cbte
        return True

    @inicializar_y_capturar_excepciones
    def AgregarItemDetalleAjuste(self, nro_item_ajustar, **kwargs):
        "Agrega el detalle de item a un ajuste de liquidación"
        d = {'nroItemAjustar': nro_item_ajustar, 'ajusteCompraAsociada': []}
        self.solicitud['itemDetalleAjusteLiquidacion'].append(d)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarAjusteFisico(self, cantidad, cantidad_cabezas=None,
                            cantidad_kg_vivo=None, **kwargs):
        "Agrega campos al detalle de item por un ajuste fisico"
        d = {'cantidad': cantidad,
             'cantidadCabezas': cantidad_cabezas,
             'cantidadKgVivo': cantidad_kg_vivo,
             }
        item_liq = self.solicitud['itemDetalleAjusteLiquidacion'][-1]
        item_liq['ajusteFisico'] = d
        return True

    @inicializar_y_capturar_excepciones
    def AgregarAjusteMonetario(self, precio_unitario, precio_recupero=None,
                               **kwargs):
        "Agrega campos al detalle de item por un ajuste monetario"
        d = {'precioUnitario': precio_unitario,
             'precioRecupero': precio_recupero,
             }
        item_liq = self.solicitud['itemDetalleAjusteLiquidacion'][-1]
        item_liq['ajusteMonetario'] = d
        return True

    @inicializar_y_capturar_excepciones
    def AgregarAjusteFinanciero(self, **kwargs):
        "Prepara el detalle de item por un ajuste financiero (gastos/tributos)"
        self.solicitud['ajusteFinanciero'] = {'gasto': [], 'tributo': []}
        return True

    @inicializar_y_capturar_excepciones
    def AjustarLiquidacion(self):
        "Generar y ajustar una liquidación para obtener del CAE"
        # limpio los elementos que no correspondan por estar vacios:
        for item_liq in self.solicitud['itemDetalleAjusteLiquidacion']:
            campo = 'ajusteCompraAsociada'
            if campo in item_liq and not item_liq[campo]:
                del item_liq[campo]
        for campo in self.solicitud.get('ajusteFinanciero', {}).copy():
            if not self.solicitud['ajusteFinanciero'][campo]:
                del self.solicitud['ajusteFinanciero'][campo]
        # llamo al webservice:
        ret = self.client.generarAjuste(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
            solicitud=self.solicitud,
        )
        # analizo la respusta
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        self.AnalizarLiquidacion(ret)
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarProvincias(self, sep="||"):
        "Consulta las provincias habilitadas"
        ret = self.client.consultarProvincias(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('provincia', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarLocalidades(self, cod_provincia, sep="||"):
        "Consulta las localidades habilitadas"
        ret = self.client.consultarLocalidadesPorProvincia(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
            solicitud={'codProvincia': cod_provincia},
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('localidad', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarOperaciones(self, sep="||"):
        "Retorna un listado de código y descripción de operaciones permitidas"
        ret = self.client.consultarOperaciones(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('operacion', []) + ret.get('operacionPorcina', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarTributos(self, sep="||"):
        "Retorna un listado de tributos con código, descripción y signo."
        ret = self.client.consultarTributos(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('tributo', []) + ret.get('tributoPorcino', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarGastos(self, sep="||"):
        "Retorna un listado de gastos con código y descripción"
        ret = self.client.consultarGastos(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('gasto', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposComprobante(self, sep="||"):
        "Retorna un listado de tipos de comprobantes con código y descripción"
        ret = self.client.consultarTiposComprobante(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('tipoComprobante', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposLiquidacion(self, sep="||"):
        "Retorna un listado de tipos de liquidación con código y descripción"
        ret = self.client.consultarTiposLiquidacion(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('tipoLiquidacion', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarCaracteres(self, sep="||"):
        "Retorna listado de caracteres emisor/receptor (código, descripción)"
        ret = self.client.consultarCaracteresParticipante(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('caracter', []) + ret.get('caracterPorcino', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarCategorias(self, sep="||"):
        "Retorna listado de categorías existentes (código, descripción)"
        ret = self.client.consultarCategorias(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('categoria', []) + ret.get('categoriaPorcina', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarMotivos(self, sep="||"):
        "Retorna listado de motivos existentes (código, descripción)"
        ret = self.client.consultarMotivos(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('motivo', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarRazas(self, sep="||"):
        "Retorna listado de razas -vacunas y porcinos- (código, descripción)"
        ret = self.client.consultarRazas(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('raza', []) + ret.get('razaPorcina', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarCortes(self, sep="||"):
        "Retorna listado de cortes -carnes- (código, descripción)"
        ret = self.client.consultarCortes(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('corte', []) + ret.get('cortePorcino', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarPuntosVentas(self, sep="||"):
        "Retorna los puntos de ventas autorizados para la utilizacion de WS"
        ret = self.client.consultarPuntosVenta(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('puntoVenta', [])
        if sep is None:
            return dict([(it['codigo'], it.get('descripcion', '')) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it.get('descripcion', '')) for it in array]

    def MostrarPDF(self, archivo, imprimir=False):
        try:
            if sys.platform == "linux2":
                os.system("evince ""%s""" % archivo)
            else:
                operation = imprimir and "print" or ""
                os.startfile(archivo, operation)
            return True
        except Exception as e:
            self.Excepcion = str(e)
            return False


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = WSLSP.InstallDir = get_install_dir()


if __name__ == '__main__':
    if '--ayuda' in sys.argv:
        print(LICENCIA)
        print(AYUDA)
        sys.exit(0)
    if '--formato' in sys.argv:
        print("Formato:")
        for msg, formato in []:
            comienzo = 1
            print("=== %s ===" % msg)
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                dec = len(fmt) > 3 and fmt[3] or (tipo == 'I' and '2' or '')
                print(" * Campo: %-20s Posición: %3d Longitud: %4d Tipo: %s Decimales: %s" % (
                    clave, comienzo, longitud, tipo, dec))
                comienzo += longitud
        sys.exit(0)

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSLSP)
        sys.exit(0)

    import csv
    from configparser import SafeConfigParser

    from .wsaa import WSAA

    try:

        if "--version" in sys.argv:
            print("Versión: ", __version__)

        if len(sys.argv) > 1 and sys.argv[1].endswith(".ini"):
            CONFIG_FILE = sys.argv[1]
            print("Usando configuracion:", CONFIG_FILE)

        config = SafeConfigParser()
        config.read(CONFIG_FILE)
        CERT = config.get('WSAA', 'CERT')
        PRIVATEKEY = config.get('WSAA', 'PRIVATEKEY')
        CUIT = config.get('WSLSP', 'CUIT')
        ENTRADA = config.get('WSLSP', 'ENTRADA')
        SALIDA = config.get('WSLSP', 'SALIDA')
        PDF = config.has_option('WSLSP', 'PDF') and config.get('WSLSP', 'PDF') or "liq.pdf"

        if config.has_option('WSAA', 'URL') and not HOMO:
            WSAA_URL = config.get('WSAA', 'URL')
        else:
            WSAA_URL = None  # wsaa.WSAAURL
        if config.has_option('WSLSP', 'URL') and not HOMO:
            WSLSP_URL = config.get('WSLSP', 'URL')
        else:
            WSLSP_URL = WSDL

        PROXY = config.has_option('WSAA', 'PROXY') and config.get('WSAA', 'PROXY') or None
        CACERT = config.has_option('WSAA', 'CACERT') and config.get('WSAA', 'CACERT') or None
        WRAPPER = config.has_option('WSAA', 'WRAPPER') and config.get('WSAA', 'WRAPPER') or None

        if config.has_section('DBF'):
            conf_dbf = dict(config.items('DBF'))
            if DEBUG:
                print("conf_dbf", conf_dbf)
        else:
            conf_dbf = {}

        DEBUG = '--debug' in sys.argv
        XML = '--xml' in sys.argv

        if DEBUG:
            print("Usando Configuración:")
            print("WSAA_URL:", WSAA_URL)
            print("WSLSP_URL:", WSLSP_URL)
            print("CACERT", CACERT)
            print("WRAPPER", WRAPPER)
        # obteniendo el TA
        from .wsaa import WSAA
        wsaa = WSAA()
        ta = wsaa.Autenticar("wslsp", CERT, PRIVATEKEY, wsdl=WSAA_URL,
                             proxy=PROXY, wrapper=WRAPPER, cacert=CACERT)
        if not ta:
            pass  # sys.exit("Imposible autenticar con WSAA: %s" % wsaa.Excepcion)

        # cliente soap del web service
        wslsp = WSLSP()
        wslsp.LanzarExcepciones = True
        wslsp.Conectar(url=WSLSP_URL, proxy=PROXY, wrapper=WRAPPER, cacert=CACERT)
        wslsp.SetTicketAcceso(ta)
        wslsp.Cuit = CUIT

        if '--dummy' in sys.argv:
            ret = wslsp.Dummy()
            print("AppServerStatus", wslsp.AppServerStatus)
            print("DbServerStatus", wslsp.DbServerStatus)
            print("AuthServerStatus", wslsp.AuthServerStatus)
            # sys.exit(0)

        if '--autorizar' in sys.argv:

            if '--prueba' in sys.argv:
                print(wslsp.client.help("generarLiquidacion"))

                # Solicitud 1: Cuenta de Venta y Líquido Producto - Hacienda
                wslsp.CrearLiquidacion(
                    cod_operacion=1,
                    fecha_cbte='2017-04-23',
                    fecha_op='2017-04-23',
                    cod_motivo=6,
                    cod_localidad_procedencia=8274,
                    cod_provincia_procedencia=1,
                    cod_localidad_destino=8274,
                    cod_provincia_destino=1,
                    lugar_realizacion='CORONEL SUAREZ',
                    fecha_recepcion=None, fecha_faena=None,
                    datos_adicionales=None)
                if False:
                    wslsp.AgregarFrigorifico(cuit=20160000156, nro_planta=1)
                wslsp.AgregarEmisor(
                    tipo_cbte=180, pto_vta=3000, nro_cbte=64,
                    cod_caracter=5, fecha_inicio_act='2016-01-01',
                    iibb='123456789', nro_ruca=305, nro_renspa=None)
                wslsp.AgregarReceptor(cod_caracter=3)
                wslsp.AgregarOperador(cuit=30160000011, iibb=3456,
                                      # nro_ruca=1011,  # Validacion AFIP 1003
                                      # cuit_autorizado=20160000261,   # 1001
                                      nro_renspa='22.123.1.12345/A4')
                wslsp.AgregarItemDetalle(
                    cuit_cliente="20160000199",     # 2403
                    cod_categoria=51020102,
                    tipo_liquidacion=1,
                    cantidad=2,
                    precio_unitario=10.0,
                    alicuota_iva=10.5,
                    cod_raza=1,
                    cantidad_cabezas=None,        # Validacion AFIP 2408
                    nro_tropa=None,
                    cod_corte=None,
                    cantidad_kg_vivo=None,
                    precio_recupero=None,
                    detalle_raza=None,
                    nro_item=1,
                )
                wslsp.AgregarCompraAsociada(tipo_cbte=185, pto_vta=3000,
                                            nro_cbte=33, cant_asoc=2,
                                            nro_item=1)
                wslsp.AgregarGuia(nro_guia=1)
                # wslsp.AgregarGuia(nro_guia=2)
                if True:
                    wslsp.AgregarDTE(nro_dte="418-1",
                                     nro_renspa='22.123.1.12345/A5')
                    wslsp.AgregarDTE(nro_dte="418-2",
                                     nro_renspa='22.123.1.12346/A5')
                else:
                    wslsp.AgregarDTE(nro_dte="418-3", nro_renspa=None)
                    wslsp.AgregarDTE(nro_dte="418-4", nro_renspa=None)
                wslsp.AgregarGasto(cod_gasto=16, base_imponible=230520.60,
                                   alicuota=3, alicuota_iva=10.5)
                wslsp.AgregarGasto(cod_gasto=99, base_imponible=None,
                                   alicuota=1, alicuota_iva=0,
                                   descripcion="Exento WSLSPv1.4.1",
                                   tipo_iva_nulo="EX")
                wslsp.AgregarTributo(cod_tributo=5, base_imponible=230520.60,
                                     alicuota=2.5)
                wslsp.AgregarTributo(cod_tributo=3, importe=397)
            else:
                # cargar un archivo de texto:
                with open(ENTRADA, "r") as f:
                    wslsp.solicitud = json.load(f, encoding="utf-8")

            if '--testing' in sys.argv:
                # mensaje de prueba (no realiza llamada remota),
                # usar solo si no está operativo, cargo respuesta:
                wslsp.LoadTestXML("tests/xml/wslsp_liq_ok_response.xml")
                import json
                with open(ENTRADA, "w") as f:
                    json.dump(wslsp.solicitud, f, sort_keys=True, indent=4, encoding="utf-8",)

            print("Liquidacion: pto_vta=%s nro_cbte=%s tipo_cbte=%s" % (
                wslsp.solicitud['emisor']['puntoVenta'],
                wslsp.solicitud['emisor']['nroComprobante'],
                wslsp.solicitud['emisor']['tipoComprobante'],
            ))

            if not '--dummy' in sys.argv:
                print("Autorizando...")
                ret = wslsp.AutorizarLiquidacion()

            if wslsp.Excepcion:
                print("EXCEPCION:", wslsp.Excepcion, file=sys.stderr)
                if DEBUG:
                    print(wslsp.Traceback, file=sys.stderr)
            print("Errores:", wslsp.Errores)
            print("CAE", wslsp.CAE)
            print("NroCodigoBarras", wslsp.NroCodigoBarras)
            print("FechaProcesoAFIP", wslsp.FechaProcesoAFIP)
            print("FechaComprobante", wslsp.FechaComprobante)
            print("NroComprobante", wslsp.NroComprobante)
            print("ImporteBruto", wslsp.ImporteBruto)
            print("ImporteTotalNeto", wslsp.ImporteTotalNeto)
            print("ImporteIVA Sobre Bruto", wslsp.ImporteIVASobreBruto)
            print("ImporteIVA Sobre Gastos", wslsp.ImporteIVASobreGastos)
            print("ImporteTotalNeto", wslsp.ImporteTotalNeto)

            pdf = wslsp.GetParametro("pdf")
            if pdf:
                open(PDF, "wb").write(pdf)

            if '--testing' in sys.argv:
                assert wslsp.CAE == "97083467167835"

            if DEBUG:
                pprint.pprint(wslsp.params_out)

        if '--ajustar' in sys.argv:
            if '--prueba' in sys.argv:
                # ejemplo documentación AFIP:
                if '--testing' in sys.argv:
                    wslsp.LoadTestXML("tests/xml/wslsp_ajuste_test.xml")
                wslsp.CrearAjuste(tipo_ajuste='C', fecha_cbte='2017-01-06',
                                  datos_adicionales='Ajuste sobre liquidacion de compra directa'
                                  )
                wslsp.AgregarEmisor(tipo_cbte=186, pto_vta=3000, nro_cbte=1)
                wslsp.AgregarComprobanteAAjustar(tipo_cbte=186, pto_vta=2000, nro_cbte=4)
                wslsp.AgregarItemDetalleAjuste(nro_item_ajustar=1)
                wslsp.AgregarCompraAsociada(tipo_cbte=185, pto_vta=3000,
                                            nro_cbte=33, cant_asoc=2,
                                            nro_item=1)
                # Validación de AFIP 3002:
                # No se pueden realizar ajustes fisicos y monetario en un mismo comprobante.
                wslsp.AgregarAjusteFisico(
                    cantidad=1,
                    cantidad_cabezas=None,
                    cantidad_kg_vivo=None,
                )
                wslsp.AgregarAjusteMonetario(
                    precio_unitario=15.995,
                    precio_recupero=None,
                )
                wslsp.AgregarAjusteFinanciero()
                wslsp.AgregarGasto(cod_gasto=16, base_imponible=230520.60,
                                   alicuota=3, alicuota_iva=10.5)
                wslsp.AgregarTributo(cod_tributo=5, base_imponible=230520.60,
                                     alicuota=2.5)
                wslsp.AgregarTributo(cod_tributo=3, importe=397)
                import json
                with open(ENTRADA, "w") as f:
                    json.dump(wslsp.solicitud, f, sort_keys=True, indent=4, encoding="utf-8",)
            else:
                # cargar un archivo de texto:
                with open(ENTRADA, "r") as f:
                    wslsp.solicitud = json.load(f, encoding="utf-8")

            wslsp.AjustarLiquidacion()
            print("CAE:", wslsp.CAE)
            print("Tipo Ajuste:", wslsp.GetParametro("tipo_ajuste"))
            print("Modo Ajuste:", wslsp.GetParametro("modo_ajuste"))
            print("Errores:", wslsp.Errores)
            if '--testing' in sys.argv:
                assert wslsp.GetParametro("cae") == "97029023118043"
                assert wslsp.GetParametro("cbte_ajuste", "tipo_cbte") == '186'
                assert wslsp.GetParametro("cbte_ajuste", "pto_vta") == '2000'
                assert wslsp.GetParametro("cbte_ajuste", "nro_cbte") == '3'

            pdf = wslsp.GetParametro("pdf")
            if pdf:
                open(PDF, "wb").write(pdf)

        if '--consultar' in sys.argv:
            tipo_cbte = 180
            pto_vta = 3000
            nro_cbte = 1
            cuit = None
            try:
                tipo_cbte = sys.argv[sys.argv.index("--consultar") + 1]
                pto_vta = sys.argv[sys.argv.index("--consultar") + 2]
                nro_cbte = sys.argv[sys.argv.index("--consultar") + 3]
                cuit = sys.argv[sys.argv.index("--consultar") + 4]
            except IndexError:
                pass
            if '--testing' in sys.argv:
                # mensaje de prueba (no realiza llamada remota),
                # usar solo si no está operativo, cargo prueba:
                wslsp.LoadTestXML("tests/xml/wslsp_cons_test.xml")
            print("Consultando: tipo_cbte=%s pto_vta=%s nro_cbte=%s" % (tipo_cbte, pto_vta, nro_cbte))
            ret = wslsp.ConsultarLiquidacion(tipo_cbte, pto_vta, nro_cbte,
                                             cuit_comprador=cuit)
            print("CAE", wslsp.CAE)
            print("Errores:", wslsp.Errores)

            if DEBUG:
                pprint.pprint(wslsp.params_out)

            if '--mostrar' in sys.argv and pdf:
                wslsp.MostrarPDF(archivo=pdf,
                                 imprimir='--imprimir' in sys.argv)

        if '--ult' in sys.argv:
            tipo_cbte = 27
            pto_vta = 1
            try:
                tipo_cbte = sys.argv[sys.argv.index("--ult") + 1]
                pto_vta = sys.argv[sys.argv.index("--ult") + 2]
            except IndexError:
                pass

            print("Consultando ultimo nro_cbte para pto_vta=%s" % pto_vta, end=' ')
            ret = wslsp.ConsultarUltimoComprobante(tipo_cbte, pto_vta)
            if wslsp.Excepcion:
                print("EXCEPCION:", wslsp.Excepcion, file=sys.stderr)
                if DEBUG:
                    print(wslsp.Traceback, file=sys.stderr)
            print("Ultimo Nro de Comprobante", wslsp.NroComprobante)
            print("Errores:", wslsp.Errores)
            sys.exit(0)

        if "--guardar" in sys.argv:
            # grabar un archivo de texto (intercambio) con el resultado:
            liq = wslsp.params_out.copy()
            if "pdf" in liq:
                del liq["pdf"]                  # eliminador binario
            with open(SALIDA, "w") as f:
                json.dump(liq, f, default=str,
                          indent=2, sort_keys=True, encoding="utf-8")

        # Recuperar parámetros:

        if '--provincias' in sys.argv:
            ret = wslsp.ConsultarProvincias()
            print("\n".join(ret))

        if '--localidades' in sys.argv:
            try:
                cod_provincia = sys.argv[sys.argv.index("--localidades") + 1]
            except BaseException:
                cod_provincia = input("Codigo Provincia:")
            ret = wslsp.ConsultarLocalidades(cod_provincia)
            print("\n".join(ret))

        if '--operaciones' in sys.argv:
            ret = wslsp.ConsultarOperaciones()
            print("\n".join(ret))

        if '--tributos' in sys.argv:
            ret = wslsp.ConsultarTributos()
            print("\n".join(ret))

        if '--gastos' in sys.argv:
            ret = wslsp.ConsultarGastos()
            print("\n".join(ret))

        if '--tipos_cbte' in sys.argv:
            ret = wslsp.ConsultarTiposComprobante()
            print("\n".join(ret))

        if '--tipos_liq' in sys.argv:
            ret = wslsp.ConsultarTiposLiquidacion()
            print("\n".join(ret))

        if '--caracteres' in sys.argv:
            ret = wslsp.ConsultarCaracteres()
            print("\n".join(ret))

        if '--categorias' in sys.argv:
            ret = wslsp.ConsultarCategorias()
            print("\n".join(ret))

        if '--motivos' in sys.argv:
            ret = wslsp.ConsultarMotivos()
            print("\n".join(ret))

        if '--razas' in sys.argv:
            ret = wslsp.ConsultarRazas()
            print("\n".join(ret))

        if '--cortes' in sys.argv:
            ret = wslsp.ConsultarCortes()
            print("\n".join(ret))

        if '--puntosventa' in sys.argv:
            ret = wslsp.ConsultarPuntosVentas()
            print("\n".join(ret))

        print("hecho.")

    except SoapFault as e:
        print("Falla SOAP:", e.faultcode, e.faultstring.encode("ascii", "ignore"), file=sys.stderr)
        sys.exit(3)
    except Exception as e:
        try:
            print(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[0], file=sys.stderr)
        except BaseException:
            print("Excepción no disponible:", type(e), file=sys.stderr)
        if DEBUG:
            raise
        sys.exit(5)
    finally:
        if XML:
            open("wslsp_request.xml", "w").write(wslsp.client.xml_request)
            open("wslsp_response.xml", "w").write(wslsp.client.xml_response)



#!usr/bin/python
# -*- coding: utf8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Régimen de información de Compras y Ventas RG3685/14 AFIP"

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2016 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.01a"

import sys
from .utils import leer, escribir, C, N, A, I, B, get_install_dir


# Diseño de registro de Importación de comprobantes de Ventas

REGINFO_CV_VENTAS_CBTE = [
    ('fecha_cbte', 8, N),
    ('tipo_cbte', 3, N),
    ('punto_vta', 5, N),
    ('cbt_desde', 20, N),
    ('cbt_hasta', 20, N),
    ('tipo_doc', 2, N),
    ('nro_doc', 20, N),
    ('nombre', 30, A),
    ('imp_total', 15, I),
    ('imp_tot_conc', 15, I),
    ('impto_liq_rni', 15, I),
    ('imp_op_ex', 15, I),
    ('impto_perc', 15, I),
    ('imp_iibb', 15, I),
    ('impto_perc_mun', 15, I),
    ('imp_internos', 15, I),
    ('moneda_id', 3, A),
    ('moneda_ctz', 10, I, 6),
    ('cant_alicuota_iva', 1, N),
    ('codigo_operacion', 1, C),
    ('imp_trib', 15, I),
    ('fecha_venc_pago', 8, A),
]

# Diseño de registro de Importación de Alícuotas de comprobantes de Ventas

REGINFO_CV_VENTAS_CBTE_ALICUOTA = [
    ('tipo_cbte', 3, N),
    ('punto_vta', 5, N),
    ('cbt_numero', 20, N),
    ('base_imp', 15, I),
    ('iva_id', 4, N),
    ('importe', 15, I),
]


if __name__ == "__main__":

    print("Usando formato registro RG3685 (regimen informativo compras/ventas)")

    if '--caea' in sys.argv:
        caea = sys.argv[sys.argv.index("--caea") + 1]
        print("Usando CAEA:", caea)
    else:
        caea = ""

    if '--serv' in sys.argv:
        fecha_serv_desde = sys.argv[sys.argv.index("--serv") + 1]
        fecha_serv_hasta = sys.argv[sys.argv.index("--serv") + 2]
        concepto = 2
    else:
        concepto = 1

    ops = {}
    for linea in open("CAB.txt"):
        reg = leer(linea, REGINFO_CV_VENTAS_CBTE)
        reg["cae"] = caea
        reg["concepto"] = concepto
        if concepto == 2:
            reg["fecha_serv_desde"] = fecha_serv_desde
            reg["fecha_serv_hasta"] = fecha_serv_hasta
        else:
            del reg['fecha_venc_pago']
        key = (reg["tipo_cbte"], reg["punto_vta"], reg["cbt_desde"])
        ops[key] = reg
        print(key)

    for linea in open("ALI.txt"):
        iva = leer(linea, REGINFO_CV_VENTAS_CBTE_ALICUOTA)
        key = (iva["tipo_cbte"], iva["punto_vta"], iva["cbt_numero"])
        reg = ops[key]
        reg["imp_neto"] = reg.get("imp_neto", 0.00) + iva["base_imp"]
        reg["imp_iva"] = reg.get("imp_iva", 0.00) + iva["importe"]
        reg.setdefault("iva", []).append(iva)

    from . import rece1
    facts = sorted(list(ops.values()),
                   key=lambda f: (f["tipo_cbte"], f["punto_vta"], f["cbt_desde"]))
    rece1.escribir_facturas(facts, open("entrada.txt", "w"))

    print("Hecho.")



#!usr/bin/python
# -*- coding: latin1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

from .rg3685 import REGINFO_CV_VENTAS_CBTE, REGINFO_CV_VENTAS_CBTE_ALICUOTA
from .utils import leer, escribir, C, N, A, I, B, get_install_dir
import traceback
import sqlite3
import unicodedata
import sys
import os
from decimal import Decimal
import datetime
import csv
"Almacenamiento de duplicados electrónicos RG1361/02 y RG1579/03 AFIP"

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2009-2015 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.22d"

LICENCIA = """
sired.py: Generador de archivos ventas para SIRED/SIAP RG1361/02 RG1579/03
(Sistema Resúmen Electrónico de Datos / Almacenamiento de Duplicados)
Copyright (C) 2009-2015 Mariano Reingart reingart@gmail.com

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para información adicional sobre garantía, soporte técnico comercial
e incorporación/distribución en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""


CUIT = '20267565393'

# ESPECIFICACIONES TECNICAS - ANEXO II RESOLUCION GENERAL N°1361
# http://www.afip.gov.ar/afip/resol136102_Anexo_II.html

categorias = {"responsable inscripto": "01",  # IVA Responsable Inscripto
              "responsable no inscripto": "02",  # IVA Responsable no Inscripto
              "no responsable": "03",  # IVA no Responsable
              "exento": "04",  # IVA Sujeto Exento
              "consumidor final": "05",  # Consumidor Final
              "monotributo": "06",  # Responsable Monotributo
              "responsable monotributo": "06",  # Responsable Monotributo
              "no categorizado": "07",  # Sujeto no Categorizado
              "importador": "08",  # Importador del Exterior
              "exterior": "09",  # Cliente del Exterior
              "liberado": "10",  # IVA Liberado Ley Nº 19.640
              "responsable inscripto - agente de percepción": "11",  # IVA Responsable Inscripto - Agente de Percepcion
              }

codigos_operacion = {
    "Z": "Exportaciones a la zona franca",
    "X": "Exportaciones al Exterior",
    "E": "Operaciones Exentas",
}

CAB_FAC_TIPO1 = [
    ('tipo_reg', 1, N),
    ('fecha_cbte', 8, N),
    ('tipo_cbte', 2, N),
    ('ctl_fiscal', 1, C),
    ('punto_vta', 4, N),
    ('cbt_numero', 8, N),
    ('cbte_nro_reg', 8, N),
    ('cant_hojas', 3, N),
    ('tipo_doc', 2, N),
    ('nro_doc', 11, N),
    ('nombre', 30, A),
    ('imp_total', 15, I),
    ('imp_tot_conc', 15, I),
    ('imp_neto', 15, I),
    ('impto_liq', 15, I),
    ('impto_liq_rni', 15, I),
    ('imp_op_ex', 15, I),
    ('impto_perc', 15, I),
    ('imp_iibb', 15, I),
    ('impto_perc_mun', 15, I),
    ('imp_internos', 15, I),
    ('transporte', 15, I),
    ('categoria', 2, N),
    ('imp_moneda_id', 3, A),
    ('imp_moneda_ctz', 10, I),
    ('alicuotas_iva', 1, N),
    ('codigo_operacion', 1, C),
    ('cae', 14, N),
    ('fecha_vto', 8, N),
    ('fecha_anulacion', 8, A),
]

# campos especiales del encabezado:
IMPORTES = ('imp_total', 'imp_tot_conc', 'imp_neto', 'impto_liq',
            'impto_liq_rni', 'imp_op_ex', 'impto_perc', 'imp_iibb',
            'impto_perc_mun', 'imp_internos')

# total
CAB_FAC_TIPO2 = [
    ('tipo_reg', 1, N),
    ('periodo', 6, N),
    ('relleno', 13, B),
    ('cant_reg_tipo_1', 8, N),
    ('relleno', 17, B),
    ('cuit', 11, N),
    ('relleno', 22, B),
    ('imp_total', 15, I),
    ('imp_tot_conc', 15, I),
    ('imp_neto', 15, I),
    ('impto_liq', 15, I),
    ('impto_liq_rni', 15, I),
    ('imp_op_ex', 15, I),
    ('impto_perc', 15, I),
    ('imp_iibb', 15, I),
    ('impto_perc_mun', 15, I),
    ('imp_internos', 15, I),
    ('relleno', 62, B),
]

DETALLE = [
    ('tipo_cbte', 2, N),
    ('ctl_fiscal', 1, C),
    ('fecha_cbte', 8, N),
    ('punto_vta', 4, N),
    ('cbt_numero', 8, N),
    ('cbte_nro_reg', 8, N),
    ('qty', 12, I),
    ('pro_umed', 2, N),
    ('pro_precio_uni', 16, I),
    ('imp_bonif', 15, I),
    ('imp_ajuste', 16, I),
    ('imp_total', 16, I),
    ('alicuota_iva', 4, I),
    ('gravado', 1, C),
    ('anulacion', 1, C),
    ('codigo', 50, A),
    ('ds', 150, A),
]

VENTAS_TIPO1 = [
    ('tipo_reg', 1, N),
    ('fecha_cbte', 8, N),
    ('tipo_cbte', 2, N),
    ('ctl_fiscal', 1, C),
    ('punto_vta', 4, N),
    ('cbt_numero', 20, N),
    ('cbte_nro_reg', 20, N),
    ('tipo_doc', 2, N),
    ('nro_doc', 11, N),
    ('nombre', 30, A),
    ('imp_total', 15, I),
    ('imp_tot_conc', 15, I),
    ('imp_neto', 15, I),
    ('alicuota_iva', 4, I),
    ('impto_liq', 15, I),
    ('impto_liq_rni', 15, I),
    ('imp_op_ex', 15, I),
    ('impto_perc', 15, I),
    ('imp_iibb', 15, I),
    ('impto_perc_mun', 15, I),
    ('imp_internos', 15, I),
    ('categoria', 2, N),
    ('imp_moneda_id', 3, A),
    ('imp_moneda_ctz', 10, I),
    ('alicuotas_iva', 1, N),
    ('codigo_operacion', 1, C),
    ('cae', 14, N),
    ('fecha_vto', 8, N),
    ('fecha_anulacion', 8, A),
    ('info_adic', 75 - 0, B),
]

VENTAS_TIPO2 = [
    ('tipo_reg', 1, N),
    ('periodo', 6, N),
    ('relleno', 29, B),
    ('cant_reg_tipo_1', 12, N),
    ('relleno', 10, B),
    ('cuit', 11, N),
    ('relleno', 30, B),
    ('imp_total', 15, I),
    ('imp_tot_conc', 15, I),
    ('imp_neto', 15, I),
    ('Relleno', 4, B),
    ('impto_liq', 15, I),
    ('impto_liq_rni', 15, I),
    ('imp_op_ex', 15, I),
    ('impto_perc', 15, I),
    ('imp_iibb', 15, I),
    ('impto_perc_mun', 15, I),
    ('imp_internos', 15, I),
    ('relleno', 122, B),
]

# Regimen de informacion de compras y ventas


def format_as_dict(format):
    return dict([(k[0], None) for k in format])


def leer_planilla(entrada, sep=','):
    "Convierte una planilla CSV a una lista de diccionarios [{'col': celda}]"

    items = []
    csv_reader = csv.reader(open(entrada), dialect='excel', delimiter=sep)
    for row in csv_reader:
        items.append(row)
    if len(items) < 2:
        raise RuntimeError('El archivo no tiene filas validos')
    if len(items[0]) < 2:
        raise RuntimeError('El archivo no tiene columnas (usar %s de separador)' % sep)
    cols = [str(it).strip() for it in items[0]]

    # armar diccionario por cada linea
    items = [dict([(cols[i], str(v).strip()) for i, v in enumerate(item)]) for item in items[1:]]

    return items


def leer_json(entrada="sired.json"):
    "Carga los datos en formato JSON [{'col': celda}]"

    import json
    items = json.load(open(entrada))

    return items


def grabar_json(salida="sired.json"):
    "Guarda los datos en formato JSON"

    import json
    json.dump(items, open(salida, "w"), sort_keys=True, indent=4)


def generar_encabezado(items):
    "Crear archivo de cabecera de facturas emitidas"

    periodo = items[0]['fecha_cbte'][:6]

    out = open("CABECERA_%s.txt" % periodo, "w")
    totales = format_as_dict(CAB_FAC_TIPO2)
    totales['periodo'] = periodo
    for key in IMPORTES:
        totales[key] = Decimal(0)

    for item in items:
        vals = format_as_dict(CAB_FAC_TIPO1)
        vals['fecha_anulacion'] = ''
        for k in list(item.keys()):
            vals[k] = item[k]
            if k in totales and k in IMPORTES:
                totales[k] = totales[k] + Decimal(item[k])
        vals['tipo_reg'] = '1'
        vals['ctl_fiscal'] = item.get('ctl_fiscal', ' ')  # C para controlador
        vals['cbte_nro_reg'] = vals['cbt_numero']
        vals['cant_hojas'] = '01'
        vals['transporte'] = '0'
        vals['categoria'] = categorias[item['categoria'].lower()]
        if vals['imp_moneda_id'] is None:
            vals['imp_moneda_id'] = 'PES'
            vals['imp_moneda_ctz'] = '1.000'
        vals['alicuotas_iva'] = max(len(item.get('ivas', [])), 1)
        if vals['codigo_operacion'] is None:
            if int(item['tipo_cbte']) in (19, 20, 21):
                vals['codigo_operacion'] = 'E'
            else:
                vals['codigo_operacion'] = ' '
        if vals['imp_tot_conc'] is None:
            vals['imp_tot_conc'] = '0'
        s = escribir(vals, CAB_FAC_TIPO1)
        out.write(s)

    totales['tipo_reg'] = '2'
    totales['cant_reg_tipo_1'] = str(len(items))
    totales['cuit'] = CUIT
    s = escribir(totales, CAB_FAC_TIPO2)
    out.write(s)
    out.close()


def generar_detalle(items):
    "Crear archivo de detalle de facturas emitidas"

    periodo = items[0]['fecha_cbte'][:6]

    out = open("DETALLE_%s.txt" % periodo, "w")

    # recorro las facturas y detalles de artículos vendidos:
    for item in items:
        for it in item.get('detalles', [{}]):
            vals = format_as_dict(DETALLE)
            # datos generales de la factura:
            vals['tipo_reg'] = '1'
            for k in ('tipo_cbte', 'fecha_cbte', 'punto_vta', 'cbt_numero'):
                vals[k] = item[k]
            vals['cbte_nro_reg'] = item['cbt_numero']  # no hay varias hojas
            vals['ctl_fiscal'] = item.get('ctl_fiscal', ' ')  # C para controlador
            vals['anulacion'] = item.get('anulacion', ' ')
            # datos del artículo:
            vals['qty'] = it.get('qty', '1')  # cantidad
            vals['pro_umed'] = it.get('umed', '07')  # unidad de medida
            vals['pro_precio_uni'] = it.get('precio', item['imp_neto'])
            vals['imp_bonif'] = it.get('bonif', '0.00')
            vals['imp_ajuste'] = it.get('ajuste', '0.00')
            vals['imp_total'] = it.get('importe', '0.00')
            # iva
            if 'iva_id' in it and it['iva_id']:
                # mapear alicuota de iva según código usado en MTX
                iva_id = int(it['iva_id'])
                if iva_id in (1, 2):
                    alicuota = None
                else:
                    alicuota = {3: "0.00", 4: "10.5", 5: "21", 6: "27"}[iva_id]
                if alicuota is None:
                    vals['gravado'] = 'E'
                else:
                    vals['gravado'] = 'G'
                vals['alicuota_iva'] = alicuota or '0.00'
            else:
                # tomar datos generales:
                vals['alicuota_iva'] = (Decimal(item['imp_total']) / Decimal(item['imp_neto']) - 1) * 100
                if float(item.get('impto_liq', item.get('imp_iva', 0))) == 0:
                    vals['gravado'] = 'E'
                else:
                    vals['gravado'] = 'G'
            # diseño libre: código de barras y descripción:
            vals['codigo'] = it.get('codigo', '')
            vals['ds'] = it.get('ds', '')
            s = escribir(vals, DETALLE)
            out.write(s)

    out.close()


def generar_ventas(items):
    "Crear archivos de ventas (registros tipo 1 y tipo 2 totales)"

    periodo = items[0]['fecha_cbte'][:6]

    out = open("VENTAS_%s.txt" % periodo, "w")
    totales = format_as_dict(VENTAS_TIPO2)
    totales['periodo'] = periodo
    for key in IMPORTES:
        totales[key] = Decimal(0)

    # recorro las facturas e itero sobre los subtotales por alicuota de IVA:
    for item in items:
        ivas = item.get("ivas", [{}])
        for i, iva in enumerate(ivas):
            vals = format_as_dict(VENTAS_TIPO1)
            # datos generales de la factura:
            vals['tipo_reg'] = '1'
            # copio los campos que no varían para las distintas alicuotas de IVA
            for k, l, t in VENTAS_TIPO1[1:10] + VENTAS_TIPO1[21:30]:
                vals[k] = item.get(k)
            vals['fecha_anulacion'] = ''
            vals['ctl_fiscal'] = item.get('ctl_fiscal', ' ')  # C para controlador
            vals['anulacion'] = item.get('anulacion', ' ')
            vals['cbte_nro_reg'] = item['cbt_numero']
            vals['cant_hojas'] = '01'
            vals['transporte'] = '0'
            vals['categoria'] = categorias[item['categoria'].lower()]
            if vals['imp_moneda_id'] is None:
                vals['imp_moneda_id'] = 'PES'
                vals['imp_moneda_ctz'] = '1.000'
            # subtotales por alícuota de IVA
            if 'iva_id' in iva:
                # mapear alicuota de iva según código usado en MTX
                iva_id = int(iva['iva_id'])
                if iva_id == 1:
                    vals['imp_tot_conc'] = iva['base_imp']
                    alicuota = None
                elif iva_id == 2:
                    vals['imp_op_ex'] = iva['base_imp']
                    alicuota = None
                else:
                    alicuota = {3: "0.00", 4: "10.5", 5: "21", 6: "27"}[iva_id]
                    vals['imp_neto'] = iva['base_imp']
                    vals['impto_liq'] = iva['importe']
                vals['alicuota_iva'] = alicuota or '0.00'
            else:
                # tomar datos generales:
                vals['alicuota_iva'] = (Decimal(item['imp_total']) / Decimal(item['imp_neto']) - 1) * 100
                vals['alicuotas_iva'] = '01'
                if float(item.get('impto_liq', item.get('imp_iva', 0))) == 0:
                    vals['codigo_operacion'] = 'E'
                else:
                    vals['codigo_operacion'] = ' '
                if vals['imp_tot_conc'] is None:
                    vals['imp_tot_conc'] = '0'

            # acumulo los totales para el registro tipo 2
            for k in IMPORTES:
                totales[k] = totales[k] + Decimal(vals[k] or 0)

            # otros impuestos (TODO: recorrer tributos) solo ultimo registro:
            if len(ivas) == i - 1:
                for k in ('impto_perc', 'imp_iibb', 'impto_perc_mun', 'imp_internos'):
                    if k in item:
                        vals[k] = item[k]
                        totales[k] = totales[k] + Decimal(vals[k] or 0)

            s = escribir(vals, VENTAS_TIPO1)
            out.write(s)

    totales['tipo_reg'] = '2'
    totales['cant_reg_tipo_1'] = str(len(items))
    totales['cuit'] = CUIT
    s = escribir(totales, VENTAS_TIPO2)
    out.write(s)
    out.close()


class SIRED():
    "Componente para Sistema Resúmen Electrónico de Datos RG1361/02 RG1579/03"

    _public_methods_ = ['CrearBD',
                        'CrearFactura',
                        'AgregarDetalleItem', 'AgregarIva', 'AgregarTributo',
                        'AgregarCmpAsoc', 'AgregarPermiso',
                        'AgregarDato',
                        'GuardarFactura', 'ObtenerFactura',
                        ]
    _public_attrs_ = ['InstallDir', 'Traceback', 'Excepcion', 'Version',
                      ]
    _readonly_attrs_ = _public_attrs_
    _reg_progid_ = "SIRED"
    _reg_clsid_ = "{3DC74AD5-939F-42AB-8381-FCA7AF783C77}"

    def __init__(self):
        self.db_path = os.path.join(self.InstallDir, "sired.db")
        self.Version = __version__
        # Abrir la base de datos
        crear = not os.path.exists(self.db_path)
        self.db = sqlite3.connect(self.db_path)
        self.db.row_factory = sqlite3.Row
        self.cursor = self.db.cursor()
        if crear:
            from .formatos.formato_txt import ENCABEZADO, DETALLE, TRIBUTO, IVA, CMP_ASOC, PERMISO, DATO
            from .formatos.formato_sql import esquema_sql
            tipos_registro = [
                ('encabezado', ENCABEZADO),
                ('detalle', DETALLE),
                ('tributo', TRIBUTO),
                ('iva', IVA),
                ('cmp_asoc', CMP_ASOC),
                ('permiso', PERMISO),
                ('dato', DATO),
            ]
            for sql in esquema_sql(tipos_registro):
                self.cursor.execute(sql)

    def CrearFactura(self, concepto=1, tipo_doc=80, nro_doc="", tipo_cbte=1, punto_vta=0,
                     cbte_nro=0, imp_total=0.00, imp_tot_conc=0.00, imp_neto=0.00,
                     imp_iva=0.00, imp_trib=0.00, imp_op_ex=0.00, fecha_cbte="", fecha_venc_pago="",
                     fecha_serv_desde=None, fecha_serv_hasta=None,
                     moneda_id="PES", moneda_ctz="1.0000", cae="", fch_venc_cae="", id_impositivo='',
                     nombre_cliente="", domicilio_cliente="", pais_dst_cmp=None,
                     obs_comerciales="", obs_generales="", forma_pago="", incoterms="",
                     idioma_cbte=7, motivos_obs="", descuento=0.0, email="",
                     **kwargs
                     ):
        "Creo un objeto factura (internamente)"
        fact = {'tipo_doc': tipo_doc, 'nro_doc': nro_doc,
                'tipo_cbte': tipo_cbte, 'punto_vta': punto_vta,
                'cbte_nro': cbte_nro,
                'imp_total': imp_total, 'imp_tot_conc': imp_tot_conc,
                'imp_neto': imp_neto, 'imp_iva': imp_iva,
                'imp_trib': imp_trib, 'imp_op_ex': imp_op_ex,
                'fecha_cbte': fecha_cbte,
                'fecha_venc_pago': fecha_venc_pago,
                'moneda_id': moneda_id, 'moneda_ctz': moneda_ctz,
                'concepto': concepto,
                'nombre_cliente': nombre_cliente,
                'domicilio_cliente': domicilio_cliente,
                'pais_dst_cmp': pais_dst_cmp,
                'obs_comerciales': obs_comerciales,
                'obs_generales': obs_generales,
                'id_impositivo': id_impositivo,
                'forma_pago': forma_pago, 'incoterms': incoterms,
                'cae': cae, 'fecha_vto': fch_venc_cae,
                'motivos_obs': motivos_obs,
                'descuento': descuento,
                'email': email,
                'cbtes_asoc': [],
                'tributos': [],
                'ivas': [],
                'permisos': [],
                'detalles': [],
                'datos': [],
                }
        if fecha_serv_desde:
            fact['fecha_serv_desde'] = fecha_serv_desde
        if fecha_serv_hasta:
            fact['fecha_serv_hasta'] = fecha_serv_hasta
        self.factura = fact
        return True

    def EstablecerParametro(self, parametro, valor):
        "Modifico un parametro general a la factura (internamente)"
        self.factura[parametro] = valor
        return True

    def AgregarDato(self, campo, valor, pagina='T'):
        "Agrego un dato a la factura (internamente)"
        self.factura["datos"].append({'campo': campo, 'valor': valor, 'pagina': pagina})
        return True

    def AgregarDetalleItem(self, u_mtx, cod_mtx, codigo, ds, qty, umed, precio,
                           bonif, iva_id, imp_iva, importe, despacho,
                           dato_a=None, dato_b=None, dato_c=None, dato_d=None, dato_e=None):
        "Agrego un item a una factura (internamente)"
        # ds = unicode(ds, "latin1") # convierto a latin1
        # Nota: no se calcula neto, iva, etc (deben venir calculados!)
        item = {
            'u_mtx': u_mtx,
            'cod_mtx': cod_mtx,
            'codigo': codigo,
            'ds': ds,
            'qty': qty,
            'umed': umed,
            'precio': precio,
            'bonif': bonif,
            'iva_id': iva_id,
            'imp_iva': imp_iva,
            'importe': importe,
            'despacho': despacho,
            'dato_a': dato_a,
            'dato_b': dato_b,
            'dato_c': dato_c,
            'dato_d': dato_d,
            'dato_e': dato_e,
        }
        self.factura['detalles'].append(item)
        return True

    def AgregarCmpAsoc(self, tipo=1, pto_vta=0, nro=0, **kwarg):
        "Agrego un comprobante asociado a una factura (interna)"
        cmp_asoc = {'cbte_tipo': tipo, 'cbte_punto_vta': pto_vta, 'cbte_nro': nro}
        self.factura['cbtes_asoc'].append(cmp_asoc)
        return True

    def AgregarTributo(self, tributo_id=0, desc="", base_imp=0.00, alic=0, importe=0.00, **kwarg):
        "Agrego un tributo a una factura (interna)"
        tributo = {'tributo_id': tributo_id, 'desc': desc, 'base_imp': base_imp,
                   'alic': alic, 'importe': importe}
        self.factura['tributos'].append(tributo)
        return True

    def AgregarIva(self, iva_id=0, base_imp=0.0, importe=0.0, **kwarg):
        "Agrego un tributo a una factura (interna)"
        iva = {'iva_id': iva_id, 'base_imp': base_imp, 'importe': importe}
        self.factura['ivas'].append(iva)
        return True

    def GuardarFactura(self):
        from .formatos.formato_sql import escribir
        escribir([self.factura], self.db)
        return self.factura['id']

    def ActualizarFactura(self, id_factura):
        from .formatos.formato_sql import modificar
        self.factura["id"] = id_factura
        modificar(self.factura, self.db)
        return True

    def ObtenerFactura(self, id_factura=None):
        from .formatos.formato_sql import leer, max_id
        if not id_factura:
            id_factura = max_id(self.db)
        facts = list(leer(self.db, ids=[id_factura]))
        if facts:
            self.factura = facts[0]
        return True

    def Consultar(self, **kwargs):
        from .formatos.formato_sql import leer
        return leer(self.db, **kwargs)


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = SIRED.InstallDir = get_install_dir()

if __name__ == '__main__':
    try:
        if hasattr(sys, "frozen") or False:
            p = os.path.dirname(os.path.abspath(sys.executable))
            os.chdir(p)
        ##sys.stdout = open("salida.txt", "a")
        entrada = {}
        for i, k in enumerate(('encabezados', 'detalles', 'ivas', 'tributos')):
            if len(sys.argv) > i + 1:
                filename = sys.argv[i + 1]
                if not filename.startswith("--") and os.path.exists(filename):
                    entrada[k] = filename
        if not entrada:
            entrada['encabezado'] = 'facturas3.csv'

        if '--prueba' in sys.argv:
            sired = SIRED()

            # creo una factura de ejemplo
            tipo_cbte = 2
            punto_vta = 4000
            fecha = datetime.datetime.now().strftime("%Y%m%d")
            concepto = 3
            tipo_doc = 80
            nro_doc = "30000000007"
            cbte_nro = 12345678
            imp_total = "122.00"
            imp_tot_conc = "3.00"
            imp_neto = "100.00"
            imp_iva = "21.00"
            imp_trib = "1.00"
            imp_op_ex = "2.00"
            imp_subtotal = "100.00"
            fecha_cbte = fecha
            fecha_venc_pago = fecha
            # Fechas del período del servicio facturado (solo si concepto = 1?)
            fecha_serv_desde = fecha
            fecha_serv_hasta = fecha
            moneda_id = 'PES'
            moneda_ctz = '1.000'
            obs_generales = "Observaciones Generales, texto libre"
            obs_comerciales = "Observaciones Comerciales, texto libre"

            nombre_cliente = 'Joao Da Silva'
            domicilio_cliente = 'Rua 76 km 34.5 Alagoas'
            pais_dst_cmp = 16
            id_impositivo = 'PJ54482221-l'
            moneda_id = '012'
            moneda_ctz = 0.5
            forma_pago = '30 dias'
            incoterms = 'FOB'
            idioma_cbte = 1
            motivo = "11"

            cae = None
            fch_venc_cae = None

            sired.CrearFactura(concepto, tipo_doc, nro_doc, tipo_cbte, punto_vta,
                               cbte_nro, imp_total, imp_tot_conc, imp_neto,
                               imp_iva, imp_trib, imp_op_ex, fecha_cbte, fecha_venc_pago,
                               fecha_serv_desde, fecha_serv_hasta,
                               moneda_id, moneda_ctz, cae, fch_venc_cae, id_impositivo,
                               nombre_cliente, domicilio_cliente, pais_dst_cmp,
                               obs_comerciales, obs_generales, forma_pago, incoterms,
                               idioma_cbte, motivo)

            tipo = 91
            pto_vta = 2
            nro = 1234
            sired.AgregarCmpAsoc(tipo, pto_vta, nro)
            tipo = 5
            pto_vta = 2
            nro = 1234
            sired.AgregarCmpAsoc(tipo, pto_vta, nro)

            tributo_id = 99
            desc = 'Impuesto Municipal Matanza'
            base_imp = "100.00"
            alic = "1.00"
            importe = "1.00"
            sired.AgregarTributo(tributo_id, desc, base_imp, alic, importe)

            iva_id = 5  # 21%
            base_imp = 100
            importe = 21
            sired.AgregarIva(iva_id, base_imp, importe)

            u_mtx = 123456
            cod_mtx = 1234567890123
            codigo = "P0001"
            ds = "Descripcion del producto P0001\n" + "Lorem ipsum sit amet " * 10
            qty = 1.00
            umed = 7
            precio = 100.00
            bonif = 0.00
            iva_id = 5
            imp_iva = 21.00
            importe = 121.00
            despacho = 'Nº 123456'
            sired.AgregarDetalleItem(u_mtx, cod_mtx, codigo, ds, qty, umed,
                                     precio, bonif, iva_id, imp_iva, importe, despacho)

            sired.AgregarDato("prueba", "1234")
            print("Prueba!")
            id_factura = sired.GuardarFactura()
            fact = sired.factura.copy()
            ok = sired.ObtenerFactura(id_factura)
            f = sired.factura

            # verificar que los datos se hayan grabado y leido correctamente:
            difs = []

            def cmp_dict(d1, d2, prefijo=None):
                global difs
                if difs is None:
                    difs = []
                for k in set(list(d1.keys()) + list(d2.keys())):
                    if k in d1 and k in d2:
                        if isinstance(d1[k], list):
                            for i, (v1, v2) in enumerate(zip(d1[k], d2[k])):
                                cmp_dict(v1, v2, (k, i))
                        else:
                            if isinstance(d1[k], Decimal) or isinstance(d2[k], Decimal):
                                d1[k] = float(d1[k])
                                d2[k] = float(d2[k])
                            if isinstance(d1[k], int) or isinstance(d2[k], int):
                                d1[k] = int(d1[k])
                                d2[k] = int(d2[k])
                            if d1[k] != d2[k]:
                                difs.append(("Dif", prefijo, k, d1[k], d2[k]))
            cmp_dict(fact, f)
            for dif in difs:
                print(dif)

            sired.EstablecerParametro("cae", "61123022925855")
            sired.EstablecerParametro("fch_venc_cae", "20110320")
            sired.EstablecerParametro("motivo_obs", "")
            ok = sired.ActualizarFactura(id_factura)
            ok = sired.ObtenerFactura(id_factura)
            assert sired.factura["cae"] == "61123022925855"

            sys.exit(0)

        if '--leer' in sys.argv:
            if '--completar_padron' in sys.argv:
                from .padron import PadronAFIP
                padron = PadronAFIP()
                padron.Conectar(trace="--trace" in sys.argv)
                from .formatos import formato_txt
                formato = formato_txt.ENCABEZADO
                categorias_iva = dict([(int(v), k) for k, v in list(categorias.items())])

            else:
                formato = VENTAS_TIPO1

            claves = [clave for clave, pos, leng in formato if clave not in ('tipo', 'info_adic')]
            csv = csv.DictWriter(open("ventas.csv", "wb"), claves, extrasaction='ignore')
            csv.writerow(dict([(k, k) for k in claves]))
            f = open("VENTAS.txt")
            for linea in f:
                if str(linea[0]) == '2':
                    datos = leer(linea, REGINFO_CV_VENTAS_CBTE)

                    if '--completar_padron' in sys.argv:
                        cuit = datos['nro_doc']
                        print("Consultando AFIP online...", cuit, end=' ')
                        ok = padron.Consultar(cuit)
                        print(padron.direccion, padron.provincia)
                        datos["nombre_cliente"] = padron.denominacion.encode("latin1")
                        datos["domicilio_cliente"] = padron.direccion.encode("latin1")
                        datos["localidad_cliente"] = "%s (CP %s) " % (
                            padron.localidad.encode("latin1"),
                            padron.cod_postal.encode("latin1"))
                        datos["provincia_cliente"] = padron.provincia.encode("latin1")
                        datos['cbte_nro'] = datos['cbt_numero_desde']
                        #datos['id_impositivo'] = categorias_iva[int(datos['categoria'])]
                    csv.writerow(datos)
            f.close()
        else:
            # cargar datos desde planillas CSV separadas o JSON:
            if entrada['encabezados'].lower().endswith("csv"):
                facturas = items = leer_planilla(entrada['encabezados'], ";")

                # pre-procesar:
                for factura in facturas:
                    for k, v in list(factura.items()):
                        # decodificar strings (evitar problemas unicode)
                        if isinstance(v, str):
                            if isinstance(v, str):
                                v = v.decode("latin1", "ignore")
                            factura[k] = unicodedata.normalize('NFKD', v).encode('ASCII', 'ignore')
                            print(k, factura[k])
                        # convertir tipos de datos desde los strings del CSV
                        if k.startswith("imp"):
                            factura[k] = float(v)
                        if k in ('cbt_desde', 'cbt_hasta', 'concepto',
                                 'punto_vta', 'tipo_cbte', 'tipo_doc',
                                 'nro_doc', 'cbt_numero'):
                            factura[k] = int(v)

                    alicuotas = {3: 0, 4: 10.5, 5: 21., 6: 27}
                    ivas = {}
                    imp_iva = 0.00

                    ruta = os.path.dirname(entrada['encabezados'])
                    prefijos = ("%(tipo_cbte)02d%(cbt_numero)08d",
                                "%(tipo_cbte)02d%(cbt_numero)06d",
                                "%(tipo_cbte)02d%(punto_vta)04d%(cbt_numero)08d",
                                )
                    for prefijo in prefijos:
                        fn = os.path.join(ruta, "%s.csv" % (prefijo % factura))
                        print("Detalle: ", fn)
                        if os.path.exists(fn):
                            det = fn
                            print("encontrado!")
                            break
                    else:
                        if 'detalles' in entrada:
                            det = entrada['detalles']
                        else:
                            det = None

                    if det:
                        detalles = leer_planilla(det, ";")

                    for det in detalles:
                        iva_id = det.get('iva_id', 5)
                        if isinstance(det.get('ds'), str):
                            det['ds'] = det['ds'].decode("latin1", "ignore")
                        if 'ds' in det:
                            det['ds'] = unicodedata.normalize('NFKD', det['ds']).encode('ASCII', 'ignore')
                        print(det)
                        if iva_id:
                            iva_id = int(iva_id)
                            if iva_id not in ivas:
                                ivas[iva_id] = {"base_imp": 0, "importe": 0, "iva_id": iva_id}

                            importe = det.get('importe', det.get('total'))
                            if importe:
                                iva = det.get('imp_iva', None)
                                importe = round(float(importe.replace(",", ".")), 2)
                                if not iva is None:
                                    iva = round(float(iva.replace(",", ".")), 2)
                                # si el iva es incorrecto o no está, liquidar:
                                if not iva and iva_id > 3:
                                    # extraer IVA incluido factura B:
                                    if factura["tipo_cbte"] in (6, 7, 8):
                                        neto = round(importe / ((100 + alicuotas[iva_id]) / 100.), 2)
                                        iva = importe - neto
                                    else:
                                        neto = importe
                                        iva = round(neto * alicuotas[iva_id] / 100., 2)
                                    print("importe iva calc:", importe, iva)
                                else:
                                    neto = importe
                                    # descontar IVA incluido factura B:
                                    if factura["tipo_cbte"] in (6, 7, 8):
                                        neto = neto - iva
                                imp_iva += iva
                                ivas[iva_id]['importe'] += iva
                                ivas[iva_id]['base_imp'] += neto
                                det['imp_iva'] = iva

                    # rearmar estructuras internas:
                    factura['detalles'] = detalles
                    factura['ivas'] = list(ivas.values())
                    factura['datos'] = []
                    factura['tributos'] = []
                    if 'imp_iva' not in factura or factura['imp_iva'] == "":
                        print("debe agregar el IVA total en el encabezado...")
                        factura['imp_iva'] = imp_iva
                    if 'cbt_numero' in factura:
                        factura['cbt_desde'] = factura['cbt_numero']
                        factura['cbt_hasta'] = factura['cbt_numero']
                    if 'nombre' in factura:
                        factura['nombre_cliente'] = factura['nombre']
                        factura['domicilio_cliente'] = factura['domicilio']
                    factura['cbte_nro'] = factura['cbt_desde']
                    if not 'concepto' in factura:
                        factura['concepto'] = 1

                    # limpio campos que no correspondan (productos vs servicios):
                    if factura['concepto'] == 1:
                        factura['fecha_venc_pago'] = None

            elif entrada['encabezados'].lower().endswith('.json'):
                items = leer_json(entrada['encabezados'])

            print("Generando encabezado...")
            generar_encabezado(items)
            print("Generando detalle...")
            generar_detalle(items)
            print("Generando ventas...")
            generar_ventas(items)
            if '--json' in sys.argv:
                grabar_json()

        print("Hecho.")
    except Exception as e:
        if '--debug' in sys.argv:
            raise
        print("Error: por favor corriga los datos y vuelva a intentar:")
        print(str(e))
        f = open("traceback.txt", "w+")
        import traceback
        traceback.print_exc(file=f)
        f.close()
    if '--debug' in sys.argv:
        input("presione enter para continuar...")
    # sys.stdout.close()



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""Módulo para Trazabilidad de Productos Médicos ANMAT - Disp. 2303/2014
según Especificación Técnica para Pruebas de Servicios (17/09/2015)"""

# Información adicional y documentación:
# http://www.sistemasagiles.com.ar/trac/wiki/TrazabilidadProductosMedicos

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2016 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.01b"

import os
import socket
import sys
import datetime
import time
import traceback
import pysimplesoap.client
from pysimplesoap.client import SoapClient, SoapFault, parse_proxy, \
    set_http_wrapper
from pysimplesoap.simplexml import SimpleXMLElement
from io import StringIO

# importo funciones compartidas:
from .utils import leer, escribir, leer_dbf, guardar_dbf, N, A, I, json, dar_nombre_campo_dbf, get_install_dir, BaseWS, inicializar_y_capturar_excepciones

HOMO = False
TYPELIB = False

WSDL = "https://servicios.pami.org.ar/trazaenprodmed.WebService?wsdl"
WSDL_PROD = "https://servicios.pami.org.ar/trazaprodmed.WebService?wsdl"


class TrazaProdMed(BaseWS):
    "Interfaz para el WebService de Trazabilidad de Productos Médicos ANMAT - PAMI - INSSJP"
    _public_methods_ = ['InformarProducto', 'CrearTransaccion',
                        'SendCancelacTransacc', 'SendCancelacTransaccParcial',
                        'GetTransaccionesWS', 'GetCatalogoElectronicoByGTIN',
                        'GetCatalogoElectronicoByGLN', 'GetMedico',
                        'Conectar', 'LeerError', 'LeerTransaccion',
                        'SetUsername', 'SetPassword',
                        'SetParametro', 'GetParametro',
                        'GetCodigoTransaccion', 'GetResultado', 'LoadTestXML']

    _public_attrs_ = [
        'Username', 'Password',
        'CodigoTransaccion', 'Errores', 'Resultado',
        'XmlRequest', 'XmlResponse',
        'Version', 'InstallDir',
        'Traceback', 'Excepcion', 'LanzarExcepciones',
        'CantPaginas', 'HayError', 'TransaccionesWS',
    ]

    _reg_progid_ = "TrazaProdMed"
    _reg_clsid_ = "{D4112556-EF2E-45D3-A2A2-7A2849A364D9}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s %s" % (__version__, HOMO and 'Homologación' or '',
                            pysimplesoap.client.__version__)

    def __init__(self, reintentos=1):
        self.Username = self.Password = None
        self.Transacciones = []
        BaseWS.__init__(self, reintentos)

    def inicializar(self):
        BaseWS.inicializar(self)
        self.CodigoTransaccion = self.Errores = self.Resultado = None
        self.Resultado = ''
        self.Errores = []   # lista de strings para la interfaz
        self.errores = []   # lista de diccionarios (uso interno)
        self.CantPaginas = self.HayError = None

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        self.errores = ret.get('errores', [])
        self.Errores = ["%s: %s" % (it['c_error'], it['d_error'])
                        for it in ret.get('errores', [])]
        self.Resultado = ret.get('resultado')

    def Conectar(self, cache=None, wsdl=None, proxy="", wrapper=None, cacert=None, timeout=30):
        # Conecto usando el método estandard:
        print("timeout", timeout)
        ok = BaseWS.Conectar(self, cache, wsdl, proxy, wrapper, cacert, timeout,
                             soap_server="jetty")

        if ok:
            # Establecer credenciales de seguridad:
            self.client['wsse:Security'] = {
                'wsse:UsernameToken': {
                    'wsse:Username': self.Username,
                    'wsse:Password': self.Password,
                }
            }
        return ok

    @inicializar_y_capturar_excepciones
    def CrearTransaccion(self,
                         f_evento, h_evento, gln_origen, gln_destino,
                         n_remito, n_factura, vencimiento, gtin, lote,
                         numero_serial, id_evento,
                         cuit_medico=None, id_obra_social=None, apellido='', nombres='',
                         tipo_documento='', n_documento='', sexo='',
                         calle='', numero='', piso='', depto='', localidad='',
                         provincia='', n_postal='', fecha_nacimiento='', telefono='',
                         nro_afiliado=None, cod_diagnostico=None, cod_hiv=None,
                         id_motivo_devolucion=None, otro_motivo_devolucion=None):
        "Inicializa internamente una estructura TransaccionDTO para informar"
        # creo la transacción con los parámetros para llamar a InformarProducto
        tx = {
            'fEvento': f_evento, 'hEvento': h_evento,
            'glnOrigen': gln_origen, 'glnDestino': gln_destino,
            'nroFactura': n_factura, 'nroRemito': n_remito,
            'vencimiento': vencimiento,
            'gtin': gtin, 'idEvento': id_evento,
            'nroSerial': numero_serial, 'lote': lote,
            'cuitMedico': cuit_medico,
            'apellidos': apellido, 'nombres': nombres, 'telefono': telefono,
            'calle': calle, 'nroCalle': numero, 'departamento': depto,
            'piso': piso, 'localidad': localidad, 'provincia': provincia,
            'codPostal': n_postal,
            'fechaNacimiento': fecha_nacimiento, 'sexo': sexo,
            'idTipoDocumento': tipo_documento, 'nroDocumento': n_documento,
            'codDiagnostico': cod_diagnostico, 'codHiv': cod_hiv,
            'idObraSocial': id_obra_social, 'nroAfiliado': nro_afiliado,
            'idMotivoDevolucion': id_motivo_devolucion,
            'otroMotivoDevolucion': otro_motivo_devolucion,
        }
        self.Transacciones.append(tx)
        return True

    @inicializar_y_capturar_excepciones
    def InformarProducto(self, usuario, password):
        "Realiza el registro de una transacción de producto. "
        # El usuario (titular del registro/distribuidor/médico/establecimiento
        # asistencial) informa el evento ocurrido para cada uno de los productos.
        res = self.client.informarProducto(
            transacciones=self.Transacciones,
            usuario=usuario,
            password=password,
        )

        ret = res['return']

        self.CodigoTransaccion = ret['codigoTransaccion']
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def SendCancelacTransacc(self, usuario, password, codigo_transaccion):
        " Realiza la cancelación de una transacción"
        res = self.client.sendCancelacTransacc(
            transaccion=codigo_transaccion,
            usuario=usuario,
            password=password,
        )

        ret = res['return']

        self.CodigoTransaccion = ret.get('codigoTransaccion')
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def SendCancelacTransaccParcial(self, usuario, password, codigo_transaccion,
                                    gtin=None, numero_serial=None):
        " Realiza la cancelación parcial de una transacción"
        res = self.client.sendCancelacTransaccParcial(
            transaccion=codigo_transaccion,
            usuario=usuario,
            password=password,
            gtin=gtin,
            serie=numero_serial,
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('codigoTransaccion')
        self.__analizar_errores(ret)
        return True

    def LeerTransaccion(self):
        "Recorro Transacciones devueltas por GetTransaccionesWS"
        # usar GetParametro para consultar el valor retornado por el webservice

        if self.Transacciones:
            # extraigo el primer item
            self.params_out = self.Transacciones.pop(0)
            return True
        else:
            # limpio los parámetros
            self.params_out = {}
            return False

    def LeerError(self):
        "Recorro los errores devueltos y devuelvo el primero si existe"

        if self.Errores:
            # extraigo el primer item
            er = self.Errores.pop(0)
            return er
        else:
            return ""

    @inicializar_y_capturar_excepciones
    def GetTransaccionesWS(self, usuario, password,
                           id_transaccion=None,
                           gln_agente_origen=None, gln_agente_destino=None,
                           gtin=None, lote=None, serie=None, id_evento=None,
                           fecha_desde_op=None, fecha_hasta_op=None,
                           fecha_desde_t=None, fecha_hasta_t=None,
                           fecha_desde_v=None, fecha_hasta_v=None,
                           n_remito=None, n_factura=None,
                           id_provincia=None, id_estado=None, nro_pag=1, offset=100,
                           ):
        "Obtiene los movimientos realizados y permite filtros de búsqueda"

        # preparo los parametros de entrada opcionales:
        kwargs = {}
        if id_transaccion is not None:
            kwargs['idTransaccion'] = id_transaccion
        if gln_agente_origen is not None:
            kwargs['glnAgenteOrigen'] = gln_agente_origen
        if gln_agente_destino is not None:
            kwargs['glnAgenteDestino'] = gln_agente_destino
        if gtin is not None:
            kwargs['gtin'] = gtin
        if lote is not None:
            kwargs['lote'] = lote
        if serie is not None:
            kwargs['serie'] = serie
        if id_evento is not None:
            kwargs['idEvento'] = id_evento
        if fecha_desde_op is not None:
            kwargs['fechaOperacionDesde'] = fecha_desde_op
        if fecha_hasta_op is not None:
            kwargs['fechaOperacionHasta'] = fecha_hasta_op
        if fecha_desde_t is not None:
            kwargs['fechaTransaccionDesde'] = fecha_desde_t
        if fecha_hasta_t is not None:
            kwargs['fechaTransaccionHasta'] = fecha_hasta_t
        if fecha_desde_v is not None:
            kwargs['fechaVencimientoDesde'] = fecha_desde_v
        if fecha_hasta_v is not None:
            kwargs['fechaVencimientoHasta'] = fecha_hasta_v
        if n_remito is not None:
            kwargs['remito'] = n_remito
        if n_factura is not None:
            kwargs['factura'] = n_factura
        if id_provincia is not None:
            kwargs['idProvincia'] = id_provincia
        if id_estado is not None:
            kwargs['idEstadoTransaccion'] = id_estado
        if nro_pag is not None:
            kwargs['pagina'] = nro_pag
        if offset is not None:
            kwargs['offset'] = offset

        # llamo al webservice
        res = self.client.getTransaccionesWS(
            usuario=usuario,
            password=password,
            **kwargs
        )
        ret = res['return']
        if ret:
            self.__analizar_errores(ret)
            self.CantPaginas = ret.get('cantPaginas')
            self.HayError = ret.get('hay_error')
            self.Transacciones = [it for it in ret.get('list', [])]
        return True

    @inicializar_y_capturar_excepciones
    def GetCatalogoElectronicoByGTIN(self, usuario, password,
                                     gtin=None, gln=None, marca=None, modelo=None,
                                     cuit=None, id_nombre_generico=None, nro_pag=1, offset=100,
                                     ):
        "Obtiene el Catálogo Electrónico de Medicamentos"

        # preparo los parametros de entrada opcionales:
        kwargs = {}
        if cuit is not None:
            kwargs['cuit'] = cuit
        if gtin is not None:
            kwargs['gtin'] = gtin
        if gln is not None:
            kwargs['gln'] = gln
        if marca is not None:
            kwargs['marca'] = marca
        if modelo is not None:
            kwargs['modelo'] = modelo
        if id_nombre_generico is not None:
            kwargs['id_nombre_generico'] = id_nombre_generico
        if nro_pag is not None:
            kwargs['pagina'] = nro_pag
        if offset is not None:
            kwargs['offset'] = offset

        # llamo al webservice
        res = self.client.getCatalogoElectronicoByGTIN(
            usuario=usuario,
            password=password,
            **kwargs
        )

        ret = res['return']
        if ret:
            self.__analizar_errores(ret)
            self.CantPaginas = ret.get('cantPaginas')
            self.HayError = ret.get('hay_error')
            self.params_out = dict([(i, it) for i, it
                                    in enumerate(ret.get('lstProductos', []))])
            return len(self.params_out)
        else:
            return 0

    def SetUsername(self, username):
        "Establezco el nombre de usuario"
        self.Username = username

    def SetPassword(self, password):
        "Establezco la contraseña"
        self.Password = password

    def GetCodigoTransaccion(self):
        "Devuelvo el código de transacción"
        return self.CodigoTransaccion

    def GetResultado(self):
        "Devuelvo el resultado"
        return self.Resultado


def main():
    "Función principal de pruebas (obtener CAE)"
    import os
    import time
    import sys
    global WSDL, LOCATION

    DEBUG = '--debug' in sys.argv

    ws = TrazaProdMed()

    ws.Username = 'testwservice'
    ws.Password = 'testwservicepsw'

    if '--prod' in sys.argv and not HOMO:
        WSDL = WSDL_PROD
        print("Usando WSDL:", WSDL)
        sys.argv.pop(sys.argv.index("--prod"))

    # Inicializo las variables y estructuras para el archivo de intercambio:
    transacciones = []
    errores = []
    formatos = []

    if '--formato' in sys.argv:
        print("Formato:")
        for msg, formato, lista in formatos:
            comienzo = 1
            print("=== %s ===" % msg)
            print("|| %-25s || %-12s || %-5s || %-4s || %-10s ||" % (
                "Nombre", "Tipo", "Long.", "Pos(txt)", "Campo(dbf)"))
            claves = []
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                clave_dbf = dar_nombre_campo_dbf(clave, claves)
                claves.append(clave_dbf)
                print("|| %-25s || %-12s || %5d ||   %4d   || %-10s ||" % (
                    clave, tipo, longitud, comienzo, clave_dbf))
                comienzo += longitud
        sys.exit(0)

    if '--cargar' in sys.argv:
        if '--dbf' in sys.argv:
            leer_dbf(formatos[:1], {})
        elif '--json' in sys.argv:
            for formato in formatos[:1]:
                archivo = open(formato[0].lower() + ".json", "r")
                d = json.load(archivo)
                formato[2].extend(d)
                archivo.close()
        else:
            for formato in formatos[:1]:
                archivo = open(formato[0].lower() + ".txt", "r")
                for linea in archivo:
                    d = leer(linea, formato[1])
                    formato[2].append(d)
                archivo.close()

    ws.Conectar("", WSDL)

    if ws.Excepcion:
        print(ws.Excepcion)
        print(ws.Traceback)
        sys.exit(-1)

    # Datos de pruebas:

    if '--test' in sys.argv:
        ws.CrearTransaccion(
            f_evento=datetime.datetime.now().strftime("%d/%m/%Y"),
            h_evento=datetime.datetime.now().strftime("%H:%M"),
            gln_origen="7791234567801", gln_destino="7791234567801",
            n_remito="R0001-12341234", n_factura="A0001-12341234",
            vencimiento=(datetime.datetime.now() + datetime.timedelta(30)).strftime("%d/%m/%Y"),
            gtin="07791234567810", lote=datetime.datetime.now().strftime("%Y"),  # R4556567
            numero_serial=int(time.time() * 10),  # A23434
            id_evento=1,
            cuit_medico="30711622507", id_obra_social=465667,
            apellido="Reingart", nombres="Mariano",
            tipo_documento="96", n_documento="28510785", sexo="M",
            calle="San Martin", numero="5656", piso="", depto="1",
            localidad="Berazategui", provincia="Buenos Aires",
            n_postal="1700", fecha_nacimiento="20/12/1972",
            telefono="5555-5555",
            nro_afiliado="9999999999999",
            cod_diagnostico="B30",
            cod_hiv="NOAP31121970",
            id_motivo_devolucion=1,
            otro_motivo_devolucion="producto fallado",
        )

    # Opciones principales:

    if '--cancela' in sys.argv:
        if '--loadxml' in sys.argv:
            ws.LoadTestXML("tests/xml/trazaprodmed_cancela_err.xml")  # cargo respuesta
        ws.SendCancelacTransacc(*sys.argv[sys.argv.index("--cancela") + 1:])
    elif '--cancela_parcial' in sys.argv:
        ws.SendCancelacTransaccParcial(*sys.argv[sys.argv.index("--cancela_parcial") + 1:])
    elif '--consulta' in sys.argv:
        ws.GetTransaccionesWS(
            *sys.argv[sys.argv.index("--consulta") + 1:]
        )
        print("CantPaginas", ws.CantPaginas)
        print("HayError", ws.HayError)
        # print "TransaccionPlainWS", ws.TransaccionPlainWS
        # parametros comunes de salida (columnas de la tabla):
        TRANSACCIONES = list(ws.Transacciones[0].keys()) if ws.Transacciones else []
        claves = [k for k in TRANSACCIONES]
        # extiendo la lista de resultado para el archivo de intercambio:
        transacciones.extend(ws.Transacciones)
        # encabezado de la tabla:
        print("||", "||".join(["%s" % clave for clave in claves]), "||")
        # recorro los datos devueltos (TransaccionPlainWS):
        while ws.LeerTransaccion():
            for clave in claves:
                print("||", ws.GetParametro(clave), end=' ')         # imprimo cada fila
            print("||")
    elif '--catalogo' in sys.argv:
        ret = ws.GetCatalogoElectronicoByGTIN(
            *sys.argv[sys.argv.index("--catalogo") + 1:]
        )
        for catalogo in list(ws.params_out.values()):
            print(catalogo)        # imprimo cada fila
    else:
        argv = [argv for argv in sys.argv if not argv.startswith("--")]
        if not transacciones:
            if len(argv) > 16:
                ws.CrearTransaccion(*argv[3:])
            else:
                print("ERROR: no se indicaron todos los parámetros requeridos")
        if ws.Transacciones:
            try:
                usuario, password = argv[1:3]
            except BaseException:
                print("ADVERTENCIA: no se indico parámetros usuario y passoword")
                usuario = password = "pruebasws"
            ws.InformarProducto(usuario, password)
            for i, tx in enumerate(transacciones):
                print("Procesando registro", i)
                tx['codigo_transaccion'] = ws.CodigoTransaccion
                errores.extend(ws.errores)
            print("|Resultado %5s|CodigoTransaccion %10s|Errores|%s|" % (
                ws.Resultado,
                ws.CodigoTransaccion,
                '|'.join(ws.Errores or []),
            ))
        else:
            print("ERROR: no se especificaron productos a informar")

    if ws.Excepcion:
        print(ws.Traceback)

    if '--grabar' in sys.argv:
        if '--dbf' in sys.argv:
            guardar_dbf(formatos, True, {})
        elif '--json' in sys.argv:
            for formato in formatos:
                archivo = open(formato[0].lower() + ".json", "w")
                json.dump(formato[2], archivo, sort_keys=True, indent=4)
                archivo.close()
        else:
            for formato in formatos:
                archivo = open(formato[0].lower() + ".txt", "w")
                for it in formato[2]:
                    archivo.write(escribir(it, formato[1]))
            archivo.close()


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = TrazaProdMed.InstallDir = get_install_dir()


if __name__ == '__main__':

    # ajusto el encoding por defecto (si se redirije la salida)
    if not hasattr(sys.stdout, "encoding") or sys.stdout.encoding is None:
        import codecs
        import locale
        sys.stdout = codecs.getwriter(locale.getpreferredencoding())(sys.stdout, "replace")
        sys.stderr = codecs.getwriter(locale.getpreferredencoding())(sys.stderr, "replace")

    if '--register' in sys.argv or '--unregister' in sys.argv:
        import pythoncom
        if TYPELIB:
            if '--register' in sys.argv:
                tlb = os.path.abspath(os.path.join(INSTALL_DIR, "typelib", "trazaprodmed.tlb"))
                print("Registering %s" % (tlb,))
                tli = pythoncom.LoadTypeLib(tlb)
                pythoncom.RegisterTypeLib(tli, tlb)
            elif '--unregister' in sys.argv:
                k = TrazaProdMed
                pythoncom.UnRegisterTypeLib(k._typelib_guid_,
                                            k._typelib_version_[0],
                                            k._typelib_version_[1],
                                            0,
                                            pythoncom.SYS_WIN32)
                print("Unregistered typelib")
        import win32com.server.register
        win32com.server.register.UseCommandLine(TrazaProdMed)
    elif "/Automate" in sys.argv:
        # MS seems to like /automate to run the class factories.
        import win32com.server.localserver
        # win32com.server.localserver.main()
        # start the server.
        win32com.server.localserver.serve([TrazaProdMed._reg_clsid_])
    else:
        main()



#!/usr/bin/python
# -*- coding: utf8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import base64
import time
import sys
import os
from .utils import json, BaseWS, inicializar_y_capturar_excepciones, get_install_dir
from . import utils
from pysimplesoap.client import SoapFault
import traceback
from .utils import date
"""Módulo para obtener Remito Electronico Carnico:
del web service WSRemCarne versión 1.0 de AFIP (RG4256/18 y RG4303/18)
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2018 Mariano Reingart"
__license__ = "LGPL 3.0"
__version__ = "1.01a"

LICENCIA = """
wsremcarne.py: Interfaz para generar Remito Electrónico Cárnico AFIP v1.0
Remito de Carnes y subproductos derivados de la faena de bovinos y porcinos
Resolución General 4256/18 y Resolución General 4303/18.
Copyright (C) 2018 Mariano Reingart reingart@gmail.com
http://www.sistemasagiles.com.ar/trac/wiki/RemitoElectronicoCarnico

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para información adicional sobre garantía, soporte técnico comercial
e incorporación/distribución en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

AYUDA = """
Opciones:
  --ayuda: este mensaje

  --debug: modo depuración (detalla y confirma las operaciones)
  --prueba: genera y autoriza una rec de prueba (no usar en producción!)
  --xml: almacena los requerimientos y respuestas XML (depuración)
  --dummy: consulta estado de servidores

  --generar: generar un remito
  --emitir: emite un remito
  --anular: anula un remito
  --autorizar: autoriza un remito

  --ult: consulta ultimo nro remito emitido
  --consultar: consulta un remito generado

  --tipos_comprobante: tabla de parametros para tipo de comprobante
  --tipos_contingencia: tipo de contingencia que puede reportar
  --tipos_categoria_emisor: tipos de categorías de emisor
  --tipos_categoria_receptor: tipos de categorías de receptor
  --tipos_estados: estados posibles en los que puede estar un remito cárnico
  --grupos_carne' grupos de los distintos tipos de cortes de carne
  --tipos_carne': tipos de corte de carne
  --codigos_domicilio: codigos de depositos habilitados para el cuit

Ver wsremcarne.ini para parámetros de configuración (URL, certificados, etc.)"
"""


# importo funciones compartidas:


# constantes de configuración (producción/homologación):

WSDL = ["https://serviciosjava.afip.gob.ar/wsremcarne/RemCarneService?wsdl",
        "https://fwshomo.afip.gov.ar/wsremcarne/RemCarneService?wsdl"]

DEBUG = False
XML = False
CONFIG_FILE = "wsremcarne.ini"
HOMO = False
ENCABEZADO = []


class WSRemCarne(BaseWS):
    "Interfaz para el WebService de Remito Electronico Carnico (Version 3)"
    _public_methods_ = ['Conectar', 'Dummy', 'SetTicketAcceso', 'DebugLog',
                        'GenerarRemito', 'EmitirRemito', 'AutorizarRemito', 'AnularRemito', 'ConsultarRemito',
                        'InformarContingencia', 'ModificarViaje', 'RegistrarRecepcion', 'ConsultarUltimoRemitoEmitido',
                        'CrearRemito', 'AgregarViaje', 'AgregarVehiculo', 'AgregarMercaderia',
                        'AgregarDatosAutorizacion', 'AgregarContingencia',
                        'ConsultarTiposCarne', 'ConsultarTiposCategoriaEmisor', 'ConsultarTiposCategoriaReceptor',
                        'ConsultarTiposComprobante', 'ConsultarTiposContingencia', 'ConsultarTiposEstado',
                        'ConsultarCodigosDomicilio', 'ConsultarGruposCarne, ConsultarPuntosEmision',
                        'SetParametros', 'SetParametro', 'GetParametro', 'AnalizarXml', 'ObtenerTagXml', 'LoadTestXML',
                        ]
    _public_attrs_ = ['XmlRequest', 'XmlResponse', 'Version', 'Traceback', 'Excepcion', 'LanzarExcepciones',
                      'Token', 'Sign', 'Cuit', 'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
                      'CodRemito', 'TipoComprobante', 'PuntoEmision',
                      'NroRemito', 'CodAutorizacion', 'FechaVencimiento', 'FechaEmision', 'Estado', 'Resultado', 'QR',
                      'ErrCode', 'ErrMsg', 'Errores', 'ErroresFormato', 'Observaciones', 'Obs', 'Evento', 'Eventos',
                      ]
    _reg_progid_ = "WSRemCarne"
    _reg_clsid_ = "{71DB0CB9-2ED7-4226-A1E6-C3FA7FB18F41}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL[HOMO]
    LanzarExcepciones = False
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')

    def Conectar(self, *args, **kwargs):
        ret = BaseWS.Conectar(self, *args, **kwargs)
        return ret

    def inicializar(self):
        self.AppServerStatus = self.DbServerStatus = self.AuthServerStatus = None
        self.CodRemito = self.TipoComprobante = self.PuntoEmision = None
        self.NroRemito = self.CodAutorizacion = self.FechaVencimiento = self.FechaEmision = None
        self.Estado = self.Resultado = self.QR = None
        self.Errores = []
        self.ErroresFormato = []
        self.Observaciones = []
        self.Eventos = []
        self.Evento = self.ErrCode = self.ErrMsg = self.Obs = ""

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        self.Errores = [err['codigoDescripcion'] for err in ret.get('arrayErrores', [])]
        self.ErroresFormato = [err['codigoDescripcionString'] for err in ret.get('arrayErroresFormato', [])]
        errores = self.Errores + self.ErroresFormato
        self.ErrCode = ' '.join(["%(codigo)s" % err for err in errores])
        self.ErrMsg = '\n'.join(["%(codigo)s: %(descripcion)s" % err for err in errores])

    def __analizar_observaciones(self, ret):
        "Comprueba y extrae observaciones si existen en la respuesta XML"
        self.Observaciones = [obs["codigoDescripcion"] for obs in ret.get('arrayObservaciones', [])]
        self.Obs = '\n'.join(["%(codigo)s: %(descripcion)s" % obs for obs in self.Observaciones])

    def __analizar_evento(self, ret):
        "Comprueba y extrae el wvento informativo si existen en la respuesta XML"
        evt = ret.get('evento')
        if evt:
            self.Eventos = [evt]
            self.Evento = "%(codigo)s: %(descripcion)s" % evt

    @inicializar_y_capturar_excepciones
    def CrearRemito(self, tipo_comprobante, punto_emision, tipo_movimiento, categoria_emisor, cuit_titular_mercaderia, cod_dom_origen,
                    tipo_receptor, categoria_receptor=None, cuit_receptor=None, cuit_depositario=None,
                    cod_dom_destino=None, cod_rem_redestinar=None, cod_remito=None, estado=None,
                    **kwargs):
        "Inicializa internamente los datos de un remito para autorizar"
        self.remito = {'tipoComprobante': tipo_comprobante, 'puntoEmision': punto_emision, 'categoriaEmisor': categoria_emisor,
                       'cuitTitularMercaderia': cuit_titular_mercaderia, 'cuitDepositario': cuit_depositario,
                       'tipoReceptor': tipo_receptor, 'categoriaReceptor': categoria_receptor, 'cuitReceptor': cuit_receptor,
                       'codDomOrigen': cod_dom_origen, 'codDomDestino': cod_dom_destino, 'tipoMovimiento': tipo_movimiento,
                       'estado': estado, 'codRemito': cod_remito,
                       'codRemRedestinado': cod_rem_redestinar,
                       'arrayMercaderias': [], 'arrayContingencias': [],
                       }
        return True

    @inicializar_y_capturar_excepciones
    def AgregarViaje(self, cuit_transportista=None, cuit_conductor=None, fecha_inicio_viaje=None, distancia_km=None, **kwargs):
        "Agrega la información referente al viaje del remito electrónico cárnico"
        self.remito['viaje'] = {'cuitTransportista': cuit_transportista,
                                'cuitConductor': cuit_conductor,
                                'fechaInicioViaje': fecha_inicio_viaje,
                                'distanciaKm': distancia_km,
                                'vehiculo': {}
                                }
        return True

    @inicializar_y_capturar_excepciones
    def AgregarVehiculo(self, dominio_vehiculo=None, dominio_acoplado=None, **kwargs):
        "Agrega la información referente al vehiculo usado en el viaje del remito electrónico cárnico"
        self.remito['viaje']['vehiculo'] = {'dominioVehiculo': dominio_vehiculo, 'dominioAcoplado': dominio_acoplado}
        return True

    @inicializar_y_capturar_excepciones
    def AgregarMercaderia(self, orden=None, cod_tipo_prod=None, cantidad=None, unidades=None, tropa=None, **kwargs):
        "Agrega la información referente a la mercadería del remito electrónico cárnico"
        mercaderia = dict(orden=orden, tropa=tropa, codTipoProd=cod_tipo_prod, cantidad=cantidad, unidadMedida=unidades)
        self.remito['arrayMercaderias'].append(dict(mercaderia=mercaderia))
        return True

    @inicializar_y_capturar_excepciones
    def AgregarDatosAutorizacion(self, nro_remito=None, cod_autorizacion=None, fecha_emision=None, fecha_vencimiento=None, **kwargs):
        "Agrega la información referente a los datos de autorización del remito electrónico cárnico"
        self.remito['datosEmision'] = dict(nroRemito=nro_remito, codAutorizacion=cod_autorizacion,
                                           fechaEmision=fecha_emision, fechaVencimiento=fecha_vencimiento,
                                           )
        return True

    @inicializar_y_capturar_excepciones
    def AgregarContingencias(self, tipo=None, observacion=None, **kwargs):
        "Agrega la información referente a los opcionales de la liq. seq."
        contingencia = dict(tipoContingencia=tipo, observacion=observacion)
        self.remito['arrayContingencias'].append(dict(contingencia=contingencia))
        return True

    @inicializar_y_capturar_excepciones
    def GenerarRemito(self, id_req, archivo="qr.png"):
        "Informar los datos necesarios para la generación de un remito nuevo"
        if not self.remito['arrayContingencias']:
            del self.remito['arrayContingencias']
        response = self.client.generarRemito(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            idReq=id_req, remito=self.remito)
        ret = response.get("generarRemitoReturn")
        if ret:
            self.__analizar_errores(ret)
            self.__analizar_observaciones(ret)
            self.__analizar_evento(ret)
            self.AnalizarRemito(ret, archivo)
        return bool(self.CodRemito)

    def AnalizarRemito(self, ret, archivo=None):
        "Extrae el resultado del remito, si existen en la respuesta XML"
        if ret:
            self.CodRemito = ret.get("codRemito")
            self.TipoComprobante = ret.get("tipoComprobante")
            self.PuntoEmision = ret.get("puntoEmision")
            datos_aut = ret.get('datosAutorizacion')
            if datos_aut:
                self.NroRemito = datos_aut.get('nroRemito')
                self.CodAutorizacion = datos_aut.get('codAutorizacion')
                self.FechaEmision = datos_aut.get('fechaEmision')
                self.FechaVencimiento = datos_aut.get('fechaVencimiento')
            self.Estado = ret.get('estado')
            self.Resultado = ret.get('resultado')
            self.QR = ret.get('qr') or ""
            if archivo:
                qr = base64.b64decode(self.QR)
                f = open(archivo, "wb")
                f.write(qr)
                f.close()

    @inicializar_y_capturar_excepciones
    def EmitirRemito(self, archivo="qr.png"):
        "Emitir Remitos que se encuentren en estado Pendiente de Emitir."
        response = self.client.emitirRemito(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            codRemito=self.remito['codRemito'],
            viaje=self.remito.get('viaje'))
        ret = response.get("emitirRemitoReturn")
        if ret:
            self.__analizar_errores(ret)
            self.__analizar_observaciones(ret)
            self.__analizar_evento(ret)
            self.AnalizarRemito(ret, archivo)
        return bool(self.CodRemito)

    @inicializar_y_capturar_excepciones
    def AutorizarRemito(self, archivo="qr.png"):
        "Autorizar o denegar un remito (cuando corresponde autorizacion) por parte del titular/depositario"
        response = self.client.autorizarRemito(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            codRemito=self.remito['codRemito'],
            estado=self.remito['estado'])
        ret = response.get("autorizarRemitoReturn")
        if ret:
            self.__analizar_errores(ret)
            self.__analizar_observaciones(ret)
            self.__analizar_evento(ret)
            self.AnalizarRemito(ret, archivo)
        return bool(self.CodRemito)

    @inicializar_y_capturar_excepciones
    def AnularRemito(self):
        "Anular un remito generado que aún no haya sido emitido"
        response = self.client.anularRemito(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            codRemito=self.remito['codRemito'])
        ret = response.get("anularRemitoReturn")
        if ret:
            self.__analizar_errores(ret)
            self.__analizar_observaciones(ret)
            self.__analizar_evento(ret)
            self.AnalizarRemito(ret)
        return bool(self.CodRemito)

    @inicializar_y_capturar_excepciones
    def ConsultarUltimoRemitoEmitido(self, tipo_comprobante=995, punto_emision=1):
        "Obtener el último número de remito que se emitió por tipo de comprobante y punto de emisión"
        response = self.client.consultarUltimoRemitoEmitido(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            tipoComprobante=tipo_comprobante,
            puntoEmision=punto_emision)
        ret = response.get("consultarUltimoRemitoReturn", {})
        id_req = ret.get("idReq", 0)
        rec = ret.get("remito", {})
        self.__analizar_errores(ret)
        self.__analizar_observaciones(ret)
        self.__analizar_evento(ret)
        self.AnalizarRemito(rec)
        return id_req

    @inicializar_y_capturar_excepciones
    def ConsultarRemito(self, cod_remito=None, id_req=None,
                        tipo_comprobante=None, punto_emision=None, nro_comprobante=None):
        "Obtener los datos de un remito generado"
        print((self.client.help("consultarRemito")))
        response = self.client.consultarRemito(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            codRemito=cod_remito,
            idReq=id_req,
            tipoComprobante=tipo_comprobante,
            puntoEmision=punto_emision,
            nroComprobante=nro_comprobante)
        ret = response.get("consultarRemitoReturn", {})
        id_req = ret.get("idReq", 0)
        self.remito = rec = ret.get("remito", {})
        self.__analizar_errores(ret)
        self.__analizar_observaciones(ret)
        self.__analizar_evento(ret)
        self.AnalizarRemito(rec)
        return id_req

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = self.client.dummy()['dummyReturn']
        self.AppServerStatus = str(results['appserver'])
        self.DbServerStatus = str(results['dbserver'])
        self.AuthServerStatus = str(results['authserver'])

    @inicializar_y_capturar_excepciones
    def ConsultarTiposComprobante(self, sep="||"):
        "Obtener el código y descripción para tipo de comprobante"
        ret = self.client.consultarTiposComprobante(
            authRequest={
                'token': self.Token, 'sign': self.Sign,
                'cuitRepresentada': self.Cuit, },
        )['consultarTiposComprobanteReturn']
        self.__analizar_errores(ret)
        array = ret.get('arrayTiposComprobante', [])
        lista = [it['codigoDescripcion'] for it in array]
        return [("%s {codigo} %s {descripcion} %s" % (sep, sep, sep)).format(**it) if sep else it for it in lista]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposContingencia(self, sep="||"):
        "Obtener el código y descripción para cada tipo de contingencia que puede reportar"
        ret = self.client.consultarTiposContingencia(
            authRequest={
                'token': self.Token, 'sign': self.Sign,
                'cuitRepresentada': self.Cuit, },
        )['consultarTiposContingenciaReturn']
        self.__analizar_errores(ret)
        array = ret.get('arrayTiposContingencia', [])
        lista = [it['codigoDescripcion'] for it in array]
        return [("%s {codigo} %s {descripcion} %s" % (sep, sep, sep)).format(**it) if sep else it for it in lista]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposCategoriaEmisor(self, sep="||"):
        "Obtener el código y descripción para tipos de categorías de emisor"
        ret = self.client.consultarTiposCategoriaEmisor(
            authRequest={
                'token': self.Token, 'sign': self.Sign,
                'cuitRepresentada': self.Cuit, },
        )['consultarCategoriasEmisorReturn']
        self.__analizar_errores(ret)
        array = ret.get('arrayCategoriasEmisor', [])
        lista = [it['codigoDescripcionString'] for it in array]
        return [("%s {codigo} %s {descripcion} %s" % (sep, sep, sep)).format(**it) if sep else it for it in lista]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposCategoriaReceptor(self, sep="||"):
        "Obtener el código y descripción para cada tipos de categorías de receptor"
        ret = self.client.consultarTiposCategoriaReceptor(
            authRequest={
                'token': self.Token, 'sign': self.Sign,
                'cuitRepresentada': self.Cuit, },
        )['consultarCategoriasReceptorReturn']
        self.__analizar_errores(ret)
        array = ret.get('arrayCategoriasReceptor', [])
        lista = [it['codigoDescripcionString'] for it in array]
        return [("%s {codigo} %s {descripcion} %s" % (sep, sep, sep)).format(**it) if sep else it for it in lista]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposEstado(self, sep="||"):
        "Obtener el código y descripción para cada estado posibles en los que puede estar un remito cárnico"
        ret = self.client.consultarTiposEstado(
            authRequest={
                'token': self.Token, 'sign': self.Sign,
                'cuitRepresentada': self.Cuit, },
        )['consultarTiposEstadoReturn']
        self.__analizar_errores(ret)
        array = ret.get('arrayTiposEstado', [])
        lista = [it['codigoDescripcionString'] for it in array]
        return [("%s {codigo} %s {descripcion} %s" % (sep, sep, sep)).format(**it) if sep else it for it in lista]

    @inicializar_y_capturar_excepciones
    def ConsultarGruposCarne(self, sep="||"):
        "Obtener el código y descripción para los grupos de los distintos tipos de cortes de carne"
        ret = self.client.consultarGruposCarne(
            authRequest={
                'token': self.Token, 'sign': self.Sign,
                'cuitRepresentada': self.Cuit, },
        )['consultarGruposCarneReturn']
        self.__analizar_errores(ret)
        array = ret.get('arrayGruposCarne', [])
        lista = [it['codigoDescripcionString'] for it in array]
        return [("%s {codigo} %s {descripcion} %s" % (sep, sep, sep)).format(**it) if sep else it for it in lista]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposCarne(self, cod_grupo_carne=1, sep="||"):
        "Obtener el código y descripción para tipos de corte de carne"
        ret = self.client.consultarTiposCarne(
            authRequest={
                'token': self.Token, 'sign': self.Sign,
                'cuitRepresentada': self.Cuit, },
            codGrupoCarne=cod_grupo_carne,
        )['consultarTiposCarneReturn']
        self.__analizar_errores(ret)
        array = ret.get('arrayTiposCarne', [])
        lista = [it['codigoDescripcionString'] for it in array]
        return [("%s {codigo} %s {descripcion} %s" % (sep, sep, sep)).format(**it) if sep else it for it in lista]

    @inicializar_y_capturar_excepciones
    def ConsultarCodigosDomicilio(self, cuit_titular=1, sep="||"):
        "Obtener el código de depositos que tiene habilitados para operar el cuit informado"
        ret = self.client.consultarCodigosDomicilio(
            authRequest={
                'token': self.Token, 'sign': self.Sign,
                'cuitRepresentada': self.Cuit, },
            cuitTitularDomicilio=cuit_titular,
        )['consultarCodigosDomicilioReturn']
        self.__analizar_errores(ret)
        array = ret.get('arrayDomicilios', [])
        lista = [it['codigoDescripcion'] for it in array]
        return [("%s {codigo} %s {descripcion} %s" % (sep, sep, sep)).format(**it) if sep else it for it in lista]


# busco el directorio de instalación (global para que no cambie si usan otra dll)
if not hasattr(sys, "frozen"):
    basepath = __file__
elif sys.frozen == 'dll':
    import win32api
    basepath = win32api.GetModuleFileName(sys.frozendllhandle)
else:
    basepath = sys.executable
INSTALL_DIR = WSRemCarne.InstallDir = get_install_dir()


if __name__ == '__main__':
    if '--ayuda' in sys.argv:
        print(LICENCIA)
        print(AYUDA)
        sys.exit(0)

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSRemCarne)
        sys.exit(0)

    from configparser import SafeConfigParser

    try:

        if "--version" in sys.argv:
            print("Versión: ", __version__)

        for arg in sys.argv[1:]:
            if arg.startswith("--"):
                break
            print("Usando configuración:", arg)
            CONFIG_FILE = arg

        config = SafeConfigParser()
        config.read(CONFIG_FILE)
        CERT = config.get('WSAA', 'CERT')
        PRIVATEKEY = config.get('WSAA', 'PRIVATEKEY')
        CUIT = config.get('WSRemCarne', 'CUIT')
        ENTRADA = config.get('WSRemCarne', 'ENTRADA')
        SALIDA = config.get('WSRemCarne', 'SALIDA')

        if config.has_option('WSAA', 'URL') and not HOMO:
            wsaa_url = config.get('WSAA', 'URL')
        else:
            wsaa_url = None
        if config.has_option('WSRemCarne', 'URL') and not HOMO:
            wsremcarne_url = config.get('WSRemCarne', 'URL')
        else:
            wsremcarne_url = WSDL[HOMO]

        if config.has_section('DBF'):
            conf_dbf = dict(config.items('DBF'))
            if DEBUG:
                print("conf_dbf", conf_dbf)
        else:
            conf_dbf = {}

        DEBUG = '--debug' in sys.argv
        XML = '--xml' in sys.argv

        if DEBUG:
            print("Usando Configuración:")
            print("wsaa_url:", wsaa_url)
            print("wsremcarne_url:", wsremcarne_url)

        # obteniendo el TA
        from .wsaa import WSAA
        wsaa = WSAA()
        ta = wsaa.Autenticar("wsremcarne", CERT, PRIVATEKEY, wsaa_url, debug=DEBUG)
        if not ta:
            sys.exit("Imposible autenticar con WSAA: %s" % wsaa.Excepcion)

        # cliente soap del web service
        wsremcarne = WSRemCarne()
        wsremcarne.Conectar(wsdl=wsremcarne_url)
        wsremcarne.SetTicketAcceso(ta)
        wsremcarne.Cuit = CUIT
        ok = None

        if '--dummy' in sys.argv:
            ret = wsremcarne.Dummy()
            print("AppServerStatus", wsremcarne.AppServerStatus)
            print("DbServerStatus", wsremcarne.DbServerStatus)
            print("AuthServerStatus", wsremcarne.AuthServerStatus)
            sys.exit(0)

        if '--ult' in sys.argv:
            try:
                pto_emision = int(sys.argv[sys.argv.index("--ult") + 1])
            except IndexError as ValueError:
                pto_emision = 1
            try:
                tipo_cbte = int(sys.argv[sys.argv.index("--ult") + 1])
            except IndexError as ValueError:
                tipo_comprobante = 995
            rec = {}
            print("Consultando ultimo remito pto_emision=%s tipo_comprobante=%s" % (pto_emision, tipo_comprobante))
            ok = wsremcarne.ConsultarUltimoRemitoEmitido(tipo_comprobante, pto_emision)
            if wsremcarne.Excepcion:
                print("EXCEPCION:", wsremcarne.Excepcion, file=sys.stderr)
                if DEBUG:
                    print(wsremcarne.Traceback, file=sys.stderr)
            print("Ultimo Nro de Remito", wsremcarne.NroRemito)
            print("Errores:", wsremcarne.Errores)

        if '--consultar' in sys.argv:
            try:
                cod_remito = sys.argv[sys.argv.index("--consultar") + 1]
            except IndexError as ValueError:
                cod_remito = None
            rec = {}
            print("Consultando remito cod_remito=%s" % (cod_remito, ))
            ok = wsremcarne.ConsultarRemito(cod_remito)
            if wsremcarne.Excepcion:
                print("EXCEPCION:", wsremcarne.Excepcion, file=sys.stderr)
                if DEBUG:
                    print(wsremcarne.Traceback, file=sys.stderr)
            print("Ultimo Nro de Remito", wsremcarne.NroRemito)
            print("Errores:", wsremcarne.Errores)
            if DEBUG:
                import pprint
                pprint.pprint(wsremcarne.remito)

        if '--prueba' in sys.argv:
            rec = dict(tipo_comprobante=995, punto_emision=1, categoria_emisor=1,
                       tipo_movimiento='ENV',  # ENV: Envio Normal, PLA: Retiro en planta, REP: Reparto, RED: Redestino
                       cuit_titular_mercaderia='20222222223', cod_dom_origen=1,
                       tipo_receptor='EM',  # 'EM': DEPOSITO EMISOR, 'MI': MERCADO INTERNO, 'RP': REPARTO
                       categoria_receptor=1, id_req=int(time.time()),
                       cuit_receptor='20111111112', cuit_depositario=None,
                       cod_dom_destino=1, cod_rem_redestinar=None,
                       cod_remito=30,
                       )
            if "--autorizar" in sys.argv:
                rec["estado"] = 'A'  # 'A': Autorizar, 'D': Denegar
            rec['viaje'] = dict(cuit_transportista='20333333334', cuit_conductor='20333333334',
                                fecha_inicio_viaje='2018-10-01', distancia_km=999)
            rec['viaje']['vehiculo'] = dict(dominio_vehiculo='AAA000', dominio_acoplado='ZZZ000')
            rec['mercaderias'] = [dict(orden=1, tropa=1, cod_tipo_prod='2.13', cantidad=10, unidades=1)]
            rec['datos_autorizacion'] = None  # dict(nro_remito=None, cod_autorizacion=None, fecha_emision=None, fecha_vencimiento=None)
            rec['contingencias'] = [dict(tipo=1, observacion="anulacion")]
            with open(ENTRADA, "w") as archivo:
                json.dump(rec, archivo, sort_keys=True, indent=4)

        if '--cargar' in sys.argv:
            with open(ENTRADA, "r") as archivo:
                rec = json.load(archivo)
            wsremcarne.CrearRemito(**rec)
            wsremcarne.AgregarViaje(**rec['viaje'])
            wsremcarne.AgregarVehiculo(**rec['viaje']['vehiculo'])
            for mercaderia in rec['mercaderias']:
                wsremcarne.AgregarMercaderia(**mercaderia)
            datos_aut = rec['datos_autorizacion']
            if datos_aut:
                wsremcarne.AgregarDatosAutorizacion(**datos_aut)
            for contingencia in rec['contingencias']:
                wsremcarne.AgregarContingencias(**contingencia)

        if '--generar' in sys.argv:
            if '--testing' in sys.argv:
                wsremcarne.LoadTestXML("tests/xml/wsremcarne_generar_response_ok_beta.xml")  # cargo respuesta

            ok = wsremcarne.GenerarRemito(id_req=rec['id_req'])

        if '--emitir' in sys.argv:
            ok = wsremcarne.EmitirRemito()

        if '--autorizar' in sys.argv:
            ok = wsremcarne.AutorizarRemito()

        if '--anular' in sys.argv:
            ok = wsremcarne.AnularRemito()

        if ok is not None:
            print("Resultado: ", wsremcarne.Resultado)
            print("Cod Remito: ", wsremcarne.CodRemito)
            if wsremcarne.CodAutorizacion:
                print("Numero Remito: ", wsremcarne.NroRemito)
                print("Cod Autorizacion: ", wsremcarne.CodAutorizacion)
                print("Fecha Emision", wsremcarne.FechaEmision)
                print("Fecha Vencimiento", wsremcarne.FechaVencimiento)
            print("Estado: ", wsremcarne.Estado)
            print("Observaciones: ", wsremcarne.Observaciones)
            print("Errores:", wsremcarne.Errores)
            print("Errores Formato:", wsremcarne.ErroresFormato)
            print("Evento:", wsremcarne.Evento)
            rec['cod_remito'] = wsremcarne.CodRemito
            rec['resultado'] = wsremcarne.Resultado
            rec['observaciones'] = wsremcarne.Observaciones
            rec['fecha_emision'] = wsremcarne.FechaEmision
            rec['fecha_vencimiento'] = wsremcarne.FechaVencimiento
            rec['errores'] = wsremcarne.Errores
            rec['errores_formato'] = wsremcarne.ErroresFormato
            rec['evento'] = wsremcarne.Evento

        if '--grabar' in sys.argv:
            with open(SALIDA, "w") as archivo:
                json.dump(rec, archivo, sort_keys=True, indent=4)

        # Recuperar parámetros:

        if '--tipos_comprobante' in sys.argv:
            ret = wsremcarne.ConsultarTiposComprobante()
            print("\n".join(ret))

        if '--tipos_contingencia' in sys.argv:
            ret = wsremcarne.ConsultarTiposContingencia()
            print("\n".join(ret))

        if '--tipos_categoria_emisor' in sys.argv:
            ret = wsremcarne.ConsultarTiposCategoriaEmisor()
            print("\n".join(ret))

        if '--tipos_categoria_receptor' in sys.argv:
            ret = wsremcarne.ConsultarTiposCategoriaReceptor()
            print("\n".join(ret))

        if '--tipos_estados' in sys.argv:
            ret = wsremcarne.ConsultarTiposEstado()
            print("\n".join(ret))

        if '--grupos_carne' in sys.argv:
            ret = wsremcarne.ConsultarGruposCarne()
            print("\n".join(ret))

        if '--tipos_carne' in sys.argv:
            for grupo_carne in wsremcarne.ConsultarGruposCarne(sep=None):
                ret = wsremcarne.ConsultarTiposCarne(grupo_carne['codigo'])
                print("\n".join(ret))

        if '--codigos_domicilio' in sys.argv:
            cuit = input("Cuit Titular Domicilio: ")
            ret = wsremcarne.ConsultarCodigosDomicilio(cuit)
            print("\n".join(ret))

        if wsremcarne.Errores or wsremcarne.ErroresFormato:
            print("Errores:", wsremcarne.Errores, wsremcarne.ErroresFormato)

        print("hecho.")

    except SoapFault as e:
        print("Falla SOAP:", e.faultcode, e.faultstring.encode("ascii", "ignore"))
        sys.exit(3)
    except Exception as e:
        ex = utils.exception_info()
        print(ex)
        if DEBUG:
            raise
        sys.exit(5)
